from copy import deepcopy

from PyQt5.QtWidgets import (
    QDialog,
    QWidget,
    QHBoxLayout,
    QPushButton,
    QVBoxLayout,
    QDoubleSpinBox,
    QGroupBox,
    QLabel,
    QComboBox,
    QStyledItemDelegate,
    QStackedWidget,
    QLineEdit,
    QFrame,
)
from PyQt5.QtCore import Qt


class BadgerIndividualLimitVariableRangeDialog(QDialog):
    def __init__(self, parent, name, apply_config, configs=None):
        super().__init__(parent)

        self.name = name
        self.apply_config = apply_config
        self.configs = deepcopy(configs)
        if configs is None:
            self.configs = {}

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        self.setWindowTitle(f"Config variable {self.name}")
        self.setMinimumWidth(360)

        vbox = QVBoxLayout(self)

        # Add the new rows above "Set range by"
        info_group = QWidget()
        vbox_info = QVBoxLayout(info_group)
        vbox_info.setContentsMargins(0, 0, 0, 0)

        # Current value row
        hbox_current = QHBoxLayout()
        lbl_current = QLabel("Current value")
        lbl_current.setFixedWidth(128)
        self.lbl_current_value = lbl_current_value = QLineEdit(
            f"{self.configs.get('current_value', 0):.2f}"
        )
        hbox_current.addWidget(lbl_current)
        hbox_current.addWidget(lbl_current_value, 1)

        # Hard lower bound row
        hbox_lower = QHBoxLayout()
        lbl_lower = QLabel("Hard lower bound")
        lbl_lower.setFixedWidth(128)
        self.lbl_hard_lower = lbl_hard_lower = QLineEdit(
            f"{self.configs.get('lower_bound', 0):.2f}"
        )
        hbox_lower.addWidget(lbl_lower)
        hbox_lower.addWidget(lbl_hard_lower, 1)

        # Hard upper bound row
        hbox_upper = QHBoxLayout()
        lbl_upper = QLabel("Hard upper bound")
        lbl_upper.setFixedWidth(128)
        self.lbl_hard_upper = lbl_hard_upper = QLineEdit(
            f"{self.configs.get('upper_bound', 0):.2f}"
        )
        hbox_upper.addWidget(lbl_upper)
        hbox_upper.addWidget(lbl_hard_upper, 1)

        # Add the rows to the info group
        vbox_info.addLayout(hbox_current)
        vbox_info.addLayout(hbox_lower)
        vbox_info.addLayout(hbox_upper)

        # Add the info group to the main layout
        vbox.addWidget(info_group)

        # Add a horizontal separator
        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)  # Horizontal line
        separator.setFrameShadow(QFrame.Sunken)  # Sunken style
        vbox.addWidget(separator)

        # Action bar
        action_bar = QWidget()
        hbox = QHBoxLayout(action_bar)
        hbox.setContentsMargins(0, 0, 0, 0)

        lbl = QLabel("Set range by")
        # lbl.setFixedWidth(128)
        self.cb = cb = QComboBox()
        tooltip = """Ratio wrt current value will set the limit
with [ratio * current value] span,
while centering around the current value;

Ratio wrt full range will set the limit
with [ratio * variable range] span,
while centering around the current value;

No matter which option you choose, the limit
will be clipped by the variable range."""
        cb.setToolTip(tooltip)
        cb.setItemDelegate(QStyledItemDelegate())
        cb.addItems(
            [
                "ratio wrt current value",
                "ratio wrt full range",
                "delta around current value",
            ]
        )
        cb.setCurrentIndex(self.configs.get("limit_option_idx", 0))

        hbox.addWidget(lbl)
        hbox.addWidget(cb, 1)

        # Config group
        group_config = QGroupBox("Parameters")
        vbox_config = QVBoxLayout(group_config)

        self.stacks = stacks = QStackedWidget()
        # Ratio curr config
        ratio_curr_config = QWidget()
        hbox_ratio_curr = QHBoxLayout(ratio_curr_config)
        hbox_ratio_curr.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel("Ratio")
        self.sb_ratio_curr = sb_ratio_curr = QDoubleSpinBox()
        sb_ratio_curr.setMinimum(0)
        sb_ratio_curr.setMaximum(1)
        sb_ratio_curr.setValue(self.configs.get("ratio_curr", 0.1))
        sb_ratio_curr.setDecimals(2)
        sb_ratio_curr.setSingleStep(0.01)
        hbox_ratio_curr.addWidget(lbl)
        hbox_ratio_curr.addWidget(sb_ratio_curr, 1)
        # Ratio full config
        ratio_full_config = QWidget()
        hbox_ratio_full = QHBoxLayout(ratio_full_config)
        hbox_ratio_full.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel("Ratio")
        self.sb_ratio_full = sb_ratio_full = QDoubleSpinBox()
        sb_ratio_full.setMinimum(0)
        sb_ratio_full.setMaximum(1)
        sb_ratio_full.setValue(self.configs.get("ratio_full", 0.1))
        sb_ratio_full.setDecimals(2)
        sb_ratio_full.setSingleStep(0.01)
        hbox_ratio_full.addWidget(lbl)
        hbox_ratio_full.addWidget(sb_ratio_full, 1)
        # Delta config
        delta_config = QWidget()
        hbox_delta = QHBoxLayout(delta_config)
        hbox_delta.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel("Delta")
        self.sb_delta = sb_delta = QDoubleSpinBox()
        sb_delta.setMinimum(0)
        sb_delta.setMaximum(1e4)
        sb_delta.setValue(self.configs.get("delta", 0.1))
        sb_delta.setDecimals(4)
        sb_delta.setSingleStep(0.01)
        hbox_delta.addWidget(lbl)
        hbox_delta.addWidget(sb_delta, 1)

        stacks.addWidget(ratio_curr_config)
        stacks.addWidget(ratio_full_config)
        stacks.addWidget(delta_config)

        stacks.setCurrentIndex(self.configs.get("limit_option_idx", 0))
        vbox_config.addWidget(stacks)
        vbox_config.addStretch(1)

        # Button set
        button_set = QWidget()
        hbox_set = QHBoxLayout(button_set)
        hbox_set.setContentsMargins(0, 0, 0, 0)
        self.btn_cancel = btn_cancel = QPushButton("Cancel")
        self.btn_set = btn_set = QPushButton("Set")
        btn_cancel.setFixedSize(96, 24)
        btn_set.setFixedSize(96, 24)
        hbox_set.addStretch()
        hbox_set.addWidget(btn_cancel)
        hbox_set.addWidget(btn_set)

        vbox.addWidget(action_bar)
        vbox.addWidget(group_config, 1)
        vbox.addWidget(button_set)

    def config_logic(self):
        self.cb.currentIndexChanged.connect(self.limit_option_changed)
        self.btn_cancel.clicked.connect(self.close)
        self.btn_set.clicked.connect(self.set)
        self.sb_ratio_curr.valueChanged.connect(self.ratio_curr_changed)
        self.sb_ratio_full.valueChanged.connect(self.ratio_full_changed)
        self.sb_delta.valueChanged.connect(self.delta_changed)

    def update_config(self):
        try:
            lower = float(self.lbl_hard_lower.text())
            upper = float(self.lbl_hard_upper.text())
            self.configs["lower_bound"] = lower
            self.configs["upper_bound"] = upper
            # Fill in default values if not exist
            if "limit_option_idx" not in self.configs:
                self.configs["limit_option_idx"] = self.cb.currentIndex()
            if "ratio_curr" not in self.configs:
                self.configs["ratio_curr"] = self.sb_ratio_curr.value()
            if "ratio_full" not in self.configs:
                self.configs["ratio_full"] = self.sb_ratio_full.value()
            if "delta" not in self.configs:
                self.configs["delta"] = self.sb_delta.value()
        except ValueError:
            pass  # Optionally handle invalid input

    def ratio_curr_changed(self, ratio_curr):
        self.configs["ratio_curr"] = ratio_curr

    def ratio_full_changed(self, ratio_full):
        self.configs["ratio_full"] = ratio_full

    def delta_changed(self, delta):
        self.configs["delta"] = delta

    def set(self):
        self.update_config()
        self.apply_config(self.name, self.configs)
        self.close()

    def limit_option_changed(self, i):
        self.stacks.setCurrentIndex(i)

        # Update configs
        self.configs["limit_option_idx"] = i

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            self.set()
        elif event.key() == Qt.Key_Escape:
            self.close()



def test_init(qtbot):
    from badger.gui.acr.components.routine_editor import BadgerRoutineEditor

    BadgerRoutineEditor()


def test_routine_set_and_save(qtbot):
    from badger.gui.acr.components.routine_editor import BadgerRoutineEditor
    from badger.tests.utils import create_routine

    window = BadgerRoutineEditor()

    routine = create_routine()
    window.set_routine(routine)

    # TODO: Test save tmp routine stuff
    # window.save_routine()



from functools import wraps
import time
from typing import Callable, Optional, cast, ParamSpec, Iterable
from types import TracebackType
from PyQt5.QtWidgets import (
    QWidget,
    QSizePolicy,
    QVBoxLayout,
    QHBoxLayout,
    QMessageBox,
    QLayout,
    QGridLayout,
    QPushButton,
    QComboBox,
    QCheckBox,
    QLabel,
    QGroupBox,
    QTabWidget,
)

from PyQt5.QtCore import Qt
from badger.gui.default.components.pf_viewer.types import PFUI, ConfigurableOptions
from badger.routine import Routine

from badger.utils import create_archive_run_filename
from matplotlib.axes import Axes
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.ticker import MaxNLocator
import pandas as pd
from torch import Tensor

from xopt.generators.bayesian.mobo import MOBOGenerator

from badger.gui.default.components.pf_viewer.types import (
    PFUIWidgets,
    PFUILayouts,
)

import logging

logger = logging.getLogger(__name__)


DEFAULT_PARAMETERS: ConfigurableOptions = {
    "plot_options": {
        "show_only_pareto_front": False,
    },
    "variable_1": 0,
    "variable_2": 1,
    "variables": [],
    "objectives": [],
    "objective_1": 0,
    "objective_2": 1,
    "plot_tab": 0,
}

Param = ParamSpec("Param")


def signal_logger(text: str):
    def decorator(fn: Callable[Param, None]) -> Callable[Param, None]:
        @wraps(fn)
        def wrapper(*args: Param.args, **kwargs: Param.kwargs):
            logger.debug(f"{text}")
            return fn(*args, **kwargs)

        return wrapper

    return decorator


class BlockSignalsContext:
    widgets: Iterable[QWidget | QLayout]

    def __init__(self, widgets: QWidget | QLayout | Iterable[QWidget | QLayout]):
        if isinstance(widgets, Iterable):
            self.widgets = widgets
        else:
            self.widgets = [widgets]

    def __enter__(self):
        for widget in self.widgets:
            if widget.signalsBlocked():
                logger.warning(
                    f"Signals already blocked for {widget} when entering context. Nesting BlockSignalsContext is not recommended as blockSignals is set to False upon exiting the context. This may lead to unexpected behavior if the widget is used again from within another BlockSignalsContext."
                )
            widget.blockSignals(True)

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ):
        for widget in self.widgets:
            if not widget.signalsBlocked():
                logger.warning(
                    f"Signals not blocked for {widget} when exiting context. Nesting BlockSignalsContext is not recommended as blockSignals is set to False upon exiting the context. This may lead to unexpected behavior if the widget is used again from within another BlockSignalsContext."
                )
            widget.blockSignals(False)


class MatplotlibFigureContext:
    def __init__(self, fig_size: tuple[float, float] | None = None):
        self.fig_size = fig_size
        self.fig = Figure(figsize=self.fig_size)
        self.ax = self.fig.add_subplot()

    def __enter__(self):
        return self.fig, self.ax

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ):
        plt.close(self.fig)


class ParetoFrontWidget(QWidget):
    routine: Routine
    generator: MOBOGenerator
    df_length: float = float("inf")
    correct_generator: bool
    last_updated: Optional[float] = None
    routine_identifier: str = ""
    hypervolume_history: pd.DataFrame = pd.DataFrame()
    pf_1: Optional[Tensor] = None
    pf_2: Optional[Tensor] = None
    pf_mask: Optional[Tensor] = None
    parameters: ConfigurableOptions = DEFAULT_PARAMETERS
    initialized: bool = False
    plot_size: tuple[float, float] = (8, 6)

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent=parent)

        self.create_ui()
        self.setWindowTitle("Pareto Front Viewer")
        self.setMinimumWidth(1200)

    def isValidRoutine(self, routine: Routine):
        if len(routine.vocs.objective_names) < 2:
            logging.error("Invalid number of objectives")
            return False

        return True

    def setup_connections(self, routine: Routine):
        self.ui["components"]["update"].clicked.connect(
            lambda: signal_logger("Update button clicked")(
                lambda: self.update_plot(routine, requires_rebuild=True)
            )()
        )

        self.ui["components"]["variables"]["variable_1"].currentIndexChanged.connect(
            lambda: signal_logger("Variable 1 has changed")(
                lambda: self.on_variable_change()
            )()
        )
        self.ui["components"]["variables"]["variable_2"].currentIndexChanged.connect(
            lambda: signal_logger("Variable 2 has changed")(
                lambda: self.on_variable_change()
            )()
        )

        self.ui["components"]["plot"]["pareto"].currentChanged.connect(
            lambda: signal_logger("Tab changed")(lambda: self.on_tab_change())()
        )

        self.ui["components"]["options"]["show_only_pareto_front"].clicked.connect(
            lambda: signal_logger("Sample checkbox changed")(lambda: self.update_ui())()
        )

    def create_ui(self):
        update_button = QPushButton("Update")
        variable_1_combo = QComboBox()
        variable_1_combo.setMinimumWidth(100)
        variable_2_combo = QComboBox()
        variable_2_combo.setMinimumWidth(100)
        show_only_pareto_front = QCheckBox("Show only Pareto Front")

        components: PFUIWidgets = {
            "variables": {
                "variable_1": variable_1_combo,
                "variable_2": variable_2_combo,
            },
            "options": {
                "show_only_pareto_front": show_only_pareto_front,
            },
            "update": update_button,
            "plot": {
                "pareto": QTabWidget(),
                "hypervolume": QVBoxLayout(),
            },
        }

        layouts: PFUILayouts = {
            "main": QHBoxLayout(),
            "settings": QVBoxLayout(),
            "plot": QGridLayout(),
            "options": QVBoxLayout(),
            "variables": QVBoxLayout(),
            "update": QVBoxLayout(),
        }

        self.ui: PFUI = {"components": components, "layouts": layouts}

        main_layout = self.ui["layouts"]["main"]

        # Left side of the layout
        settings_layout = self.ui["layouts"]["settings"]

        settings_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        # Variables layout

        variables_layout = self.ui["layouts"]["variables"]

        variable_1_layout = QHBoxLayout()
        variable_1_layout.addWidget(QLabel("X Axis"))
        variable_1_layout.addWidget(variable_1_combo)

        variable_2_layout = QHBoxLayout()
        variable_2_layout.addWidget(
            QLabel("Y Axis"), alignment=Qt.AlignmentFlag.AlignLeft
        )
        variable_2_layout.addWidget(variable_2_combo)

        variables_layout.addLayout(variable_1_layout)
        variables_layout.addLayout(variable_2_layout)

        variables_group = QGroupBox("Variables")
        variables_group.setLayout(variables_layout)

        settings_layout.addWidget(variables_group)

        # Options layout
        options_layout = self.ui["layouts"]["options"]

        show_only_pareto_front = self.ui["components"]["options"][
            "show_only_pareto_front"
        ]
        show_only_pareto_front.setChecked(
            self.parameters["plot_options"]["show_only_pareto_front"]
        )

        options_layout.addWidget(show_only_pareto_front)

        settings_layout.addLayout(options_layout)

        # Update layout

        update_button = self.ui["components"]["update"]

        settings_layout.addStretch(1)
        settings_layout.addWidget(update_button)
        settings_widget = QWidget()
        settings_widget.setLayout(settings_layout)
        settings_widget.setSizePolicy(
            QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Expanding
        )
        main_layout.addWidget(settings_widget)

        # Right side of the layout
        plot_layout = self.ui["layouts"]["plot"]
        plot_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        plot_tab_widget = self.ui["components"]["plot"]["pareto"]
        plot_tab_widget.setCurrentIndex(self.parameters["plot_tab"])
        plot_tab_widget.setMinimumWidth(400)

        plot_hypervolume_widget = QWidget()
        plot_hypervolume = self.ui["components"]["plot"]["hypervolume"]
        plot_hypervolume_widget.setLayout(plot_hypervolume)
        plot_hypervolume_widget.setMinimumWidth(400)
        plot_hypervolume_widget.setMaximumWidth(600)

        plot_layout.addWidget(plot_tab_widget, 0, 0, Qt.AlignmentFlag.AlignCenter)
        plot_layout.addWidget(
            plot_hypervolume_widget, 0, 1, Qt.AlignmentFlag.AlignCenter
        )

        main_layout.addLayout(plot_layout, 1)

        self.setLayout(main_layout)

    def initialize_ui(self):
        routine = self.routine

        # Setup the variable dropdowns
        variable_names = routine.vocs.variable_names
        objective_names = routine.vocs.objective_names

        self.parameters["variables"] = variable_names
        self.parameters["objectives"] = objective_names

        variable_1_combo = self.ui["components"]["variables"]["variable_1"]
        variable_2_combo = self.ui["components"]["variables"]["variable_2"]

        with BlockSignalsContext([variable_1_combo, variable_2_combo]):
            variable_1_combo.clear()
            variable_2_combo.clear()

            for variable_name in variable_names:
                variable_1_combo.addItem(variable_name)
                variable_2_combo.addItem(variable_name)

            variable_1_combo.setCurrentIndex(self.parameters["variable_1"])
            variable_2_combo.setCurrentIndex(self.parameters["variable_2"])

    def on_tab_change(self):
        self.parameters["plot_tab"] = self.ui["components"]["plot"][
            "pareto"
        ].currentIndex()
        # change x and y axis options
        x_combo = self.ui["components"]["variables"]["variable_1"]
        y_combo = self.ui["components"]["variables"]["variable_2"]

        plot_tab = self.parameters["plot_tab"]

        if plot_tab in (0, 1):
            with BlockSignalsContext([x_combo, y_combo]):
                x_combo.clear()
                y_combo.clear()

                if plot_tab == 0:
                    for variable_name in self.parameters["variables"]:
                        x_combo.addItem(variable_name)
                        y_combo.addItem(variable_name)

                    x_combo.setCurrentIndex(self.parameters["variable_1"])
                    y_combo.setCurrentIndex(self.parameters["variable_2"])
                elif plot_tab == 1:
                    for variable_name in self.parameters["objectives"]:
                        x_combo.addItem(variable_name)
                        y_combo.addItem(variable_name)
                    x_combo.setCurrentIndex(self.parameters["objective_1"])
                    y_combo.setCurrentIndex(self.parameters["objective_2"])
        else:
            logging.error("Invalid plot tab")
            raise ValueError("Invalid plot tab")

        # Update the plot
        self.update_pareto_front_plot()

    def on_variable_change(self):
        plot_tab = self.parameters["plot_tab"]
        if plot_tab == 0:
            self.parameters["variable_1"] = self.ui["components"]["variables"][
                "variable_1"
            ].currentIndex()
            self.parameters["variable_2"] = self.ui["components"]["variables"][
                "variable_2"
            ].currentIndex()
        elif plot_tab == 1:
            self.parameters["objective_1"] = self.ui["components"]["variables"][
                "variable_1"
            ].currentIndex()
            self.parameters["objective_2"] = self.ui["components"]["variables"][
                "variable_2"
            ].currentIndex()
        else:
            logging.error("Invalid plot tab")
            raise ValueError("Invalid plot tab")

        self.update_pareto_front_plot()

    def update_ui(self):
        self.create_plots(requires_rebuild=True)

    def clear_tabs(self, tab_widget: QTabWidget):
        max_index = tab_widget.count()
        for i in range(max_index - 1, -1, -1):
            tab_widget.removeTab(i)

    def update_pareto_front_plot(
        self,
    ):
        self.update_pareto_front()

        plot_tab_widget = self.ui["components"]["plot"]["pareto"]

        with BlockSignalsContext(plot_tab_widget):
            self.clear_tabs(plot_tab_widget)

            with MatplotlibFigureContext(self.plot_size) as (fig, ax):
                try:
                    fig0, ax0 = self.create_pareto_plot(fig, ax)
                    canvas0 = FigureCanvas(fig0)

                    ax0.set_title("Data Points")
                    plot_tab_widget.addTab(canvas0, "Variable Space")
                except ValueError:
                    logging.error("No data points available for Variable Space")
                    blank_canvas = FigureCanvas(fig)
                    plot_tab_widget.addTab(blank_canvas, "Variable Space")

            with MatplotlibFigureContext(self.plot_size) as (fig, ax):
                try:
                    fig1, ax1 = self.create_pareto_plot(fig, ax)
                    canvas1 = FigureCanvas(fig1)

                    ax1.set_title("Pareto Front")
                    plot_tab_widget.addTab(canvas1, "Objective Space")
                except ValueError:
                    logging.error("No data points available for Objective Space")
                    blank_canvas = FigureCanvas(fig)
                    plot_tab_widget.addTab(blank_canvas, "Objective Space")

            plot_tab_widget.setCurrentIndex(self.parameters["plot_tab"])

    def update_hypervolume_plot(
        self,
    ):
        self.update_hypervolume()

        plot_hypervolume = self.ui["components"]["plot"]["hypervolume"]

        with BlockSignalsContext(plot_hypervolume):
            self.clearLayout(plot_hypervolume)

            with MatplotlibFigureContext(self.plot_size) as (fig, ax):
                try:
                    fig0, ax = self.create_hypervolume_plot(fig, ax)

                    canvas = FigureCanvas(fig0)
                    plot_hypervolume.addWidget(canvas)
                except ValueError:
                    logging.error("No data points available for Hypervolume")
                    blank_canvas = FigureCanvas(fig)
                    plot_hypervolume.addWidget(blank_canvas)

    def create_plots(
        self,
        requires_rebuild: bool = False,
        interval: int = 1000,
    ):
        # Check if the plot was updated recently
        if self.last_updated is not None and not requires_rebuild:
            logger.debug(f"Time since last update: {time.time() - self.last_updated}")

            time_diff = time.time() - self.last_updated

            # If the plot was updated less than 1 second ago, skip this update
            if time_diff < interval / 1000:
                logger.debug("Skipping update")
                return

        self.update_pareto_front_plot()

        self.update_hypervolume_plot()

        # Update the last updated time
        self.last_updated = time.time()

    def create_pareto_plot(self, fig: Figure, ax: Axes):
        current_tab = self.parameters["plot_tab"]
        show_only_pareto_front = self.ui["components"]["options"][
            "show_only_pareto_front"
        ].isChecked()

        if current_tab == 0:
            x_axis = self.parameters["variable_1"]
            y_axis = self.parameters["variable_2"]

            x_var_name = self.generator.vocs.variable_names[x_axis]
            x_var_index = self.generator.vocs.variable_names.index(x_var_name)
            y_var_name = self.generator.vocs.variable_names[y_axis]
            y_var_index = self.generator.vocs.variable_names.index(y_var_name)

        elif current_tab == 1:
            x_axis = self.parameters["objective_1"]
            y_axis = self.parameters["objective_2"]

            x_var_name = self.generator.vocs.objective_names[x_axis]
            x_var_index = self.generator.vocs.objective_names.index(x_var_name)
            y_var_name = self.generator.vocs.objective_names[y_axis]
            y_var_index = self.generator.vocs.objective_names.index(y_var_name)
        else:
            logging.error("Invalid plot index")
            raise ValueError("Invalid plot index")

        if self.pf_mask is None or self.pf_1 is None or self.pf_2 is None:
            logging.error("No pareto front")
            raise ValueError("No pareto front")

        raw_data = self.generator.data

        if raw_data is None or len(raw_data) == 0:
            logging.error("No raw data available")
            raise ValueError("No raw data available")

        if current_tab == 0:
            data_points = self.pf_1
        elif current_tab == 1:
            data_points = self.pf_2
        else:
            logging.error("Invalid plot index")
            raise ValueError("Invalid plot index")

        data_indices: list[int] = [
            x for x in range(len(self.pf_mask)) if self.pf_mask[x]
        ]

        # Color bar

        num_of_points = len(raw_data)

        color_map = plt.cm.get_cmap("viridis")
        norm = Normalize(0, num_of_points - 1)

        # Map the indices to the colormap, normalizing by the total number of points
        if num_of_points > 1:
            colors = [color_map(i / (num_of_points - 1)) for i in range(num_of_points)]
        else:
            colors = [color_map(0)]

        if not show_only_pareto_front:
            x = raw_data[f"{x_var_name}"]
            y = raw_data[f"{y_var_name}"]

            if len(x) > 0 and len(y) > 0:
                ax.scatter(
                    x,
                    y,
                    c=colors,
                )

        pf_colors = [colors[i] for i in data_indices]

        ax.scatter(
            data_points[:, x_var_index],
            data_points[:, y_var_index],
            c=pf_colors,
        )

        mappable = plt.cm.ScalarMappable(cmap=color_map, norm=norm)

        fig.colorbar(
            mappable,
            ax=ax,
            orientation="vertical",
            label="Iterations",
        )

        ax.set_xlabel(x_var_name)
        ax.set_ylabel(y_var_name)

        return fig, ax

    def create_hypervolume_plot(self, fig: Figure, ax: Axes):
        data_points = self.hypervolume_history
        if len(data_points) == 0:
            raise ValueError("No data points available")
        # Extract the x and y coordinates from the data points
        x = data_points["iteration"].values
        y = data_points["hypervolume"].values
        # Create a scatter plot
        ax.plot(x, y, color="black")
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        # Set the x and y axis labels
        ax.set_xlabel("Iterations")
        ax.set_ylabel("Hypervolume")
        # Set the title of the plot
        ax.set_title("Hypervolume")

        return fig, ax

    def clearLayout(self, layout: QLayout):
        while layout.count():
            child = layout.takeAt(0)
            if child is None:
                break

            widget = child.widget()
            if widget is None:
                break

            widget.deleteLater()

    def requires_reinitialization(self):
        # Check if the extension needs to be reinitialized
        logger.debug("Checking if extension needs to be reinitialized")

        archive_name = create_archive_run_filename(self.routine)

        if not self.initialized:
            logger.debug("Reset - Extension never initialized")
            self.initialized = True
            self.routine_identifier = archive_name
            self.hypervolume_history = pd.DataFrame()
            self.pf_1 = None
            self.pf_2 = None
            self.setup_connections(self.routine)
            return True

        if self.routine_identifier != archive_name:
            logger.debug("Reset - Routine name has changed")
            self.routine_identifier = archive_name
            self.hypervolume_history = pd.DataFrame()
            self.pf_1 = None
            self.pf_2 = None
            self.pf_mask = None
            return True

        if self.routine.data is None:
            logger.debug("Reset - No data available")
            self.hypervolume_history = pd.DataFrame()
            self.pf_1 = None
            self.pf_2 = None
            self.pf_mask = None
            return True

        previous_len = self.df_length
        self.df_length = len(self.routine.data)
        new_length = self.df_length

        if previous_len > new_length:
            logger.debug("Reset - Data length is smaller")
            self.hypervolume_history = pd.DataFrame()
            self.pf_1 = None
            self.pf_2 = None
            self.pf_mask = None
            self.df_length = float("inf")
            return True

        return False

    def update_plot(self, routine: Routine, requires_rebuild: bool = False):
        if not self.update_routine(routine):
            logging.error("Failed to update routine")
            return

        if not self.isValidRoutine(self.routine):
            logging.error("Invalid routine")
            return

        if self.requires_reinitialization():
            self.initialize_ui()

        self.create_plots(requires_rebuild)

    def update_hypervolume(self):
        # Get the hypervolume from the generator
        self.generator.update_pareto_front_history()
        pareto_front_history_df = self.generator.pareto_front_history
        if pareto_front_history_df is None or len(pareto_front_history_df) == 0:
            logger.error("No hypervolume data available")
            return

        self.hypervolume_history = pareto_front_history_df

    def update_pareto_front(self):
        pf_1, pf_2, pf_mask, _ = self.generator.get_pareto_front_and_hypervolume()

        if pf_mask is None or pf_1 is None or pf_2 is None:
            logging.error("No pareto front")
            return

        self.pf_mask = pf_mask[1:]
        self.pf_1 = pf_1
        self.pf_2 = pf_2

    def update_routine(self, routine: Routine):
        logger.debug("Updating routine in Pareto Front Viewer")
        is_success = False

        self.routine = routine

        # Check if the generator is a BayesianGenerator
        if not issubclass(self.routine.generator.__class__, MOBOGenerator):
            self.correct_generator = False
            QMessageBox.critical(
                self,
                "Invalid Generator",
                f"Invalid generator type: {type(self.routine.generator)}, BO Visualizer only supports BayesianGenerator",
            )
            raise TypeError(
                f"Invalid generator type: {type(self.routine.generator)}, BO Visualizer only supports MOBOGenerator"
            )

        self.correct_generator = True

        generator = cast(MOBOGenerator, self.routine.generator)

        if generator.data is None:
            logger.error("No data available in generator")
            return is_success

        self.generator = generator
        self.df_length = len(generator.data)
        is_success = True
        return is_success



from abc import ABC, abstractmethod
from typing import List


class Extension(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def __init__(self):
        pass

    # List all available generators
    @abstractmethod
    def list_generator(self) -> List[str]:
        pass

    # Get config of an generator
    @abstractmethod
    def get_generator_config(self, name: str):
        pass

    # Run an optimization with an array-style evaluate function
    # and a configs dict
    @abstractmethod
    def optimize(self, evaluate, configs: dict):
        pass



from PyQt5.QtWidgets import QDoubleSpinBox, QAbstractSpinBox
from PyQt5.QtCore import Qt
from badger.gui.default.utils import MouseWheelWidgetAdjustmentGuard


class RobustSpinBox(QDoubleSpinBox):
    def __init__(self, *args, **kwargs):
        try:
            decimals = kwargs["decimals"]
            del kwargs["decimals"]
        except:
            decimals = 4

        try:
            lb = kwargs["lower_bound"]
            del kwargs["lower_bound"]
            if lb is None:
                lb = -1e3
        except:
            lb = -1e3

        try:
            ub = kwargs["upper_bound"]
            del kwargs["upper_bound"]
            if ub is None:
                ub = 1e3
        except:
            ub = 1e3

        try:
            default_value = kwargs["default_value"]
            del kwargs["default_value"]
            if default_value is None:
                default_value = 0
        except:
            default_value = 0

        super().__init__(*args, **kwargs)

        self.setDecimals(decimals)
        self.setFocusPolicy(Qt.StrongFocus)
        self.installEventFilter(MouseWheelWidgetAdjustmentGuard(self))
        self.setRange(lb, ub)
        self.setStepType(QAbstractSpinBox.AdaptiveDecimalStepType)
        self.setValue(default_value)



from importlib import resources
from PyQt5.QtWidgets import QHBoxLayout, QWidget, QPushButton
from PyQt5.QtGui import QIcon
from PyQt5.QtCore import Qt, QSize
from badger.gui.default.windows.settings_dialog import BadgerSettingsDialog
from badger.gui.default.components.eliding_label import SimpleElidedLabel


class BadgerStatusBar(QWidget):
    def __init__(self):
        super().__init__()

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        icon_ref = resources.files(__package__) / "../images/gear.png"
        with resources.as_file(icon_ref) as icon_path:
            self.icon_settings = QIcon(str(icon_path))

        hbox = QHBoxLayout(self)
        hbox.setContentsMargins(0, 0, 0, 0)

        self.summary = summary = SimpleElidedLabel()
        # summary.setStyleSheet("background-color: orange;")  # for debugging
        summary.setAlignment(Qt.AlignCenter)

        self.btn_settings = btn_settings = QPushButton()
        btn_settings.setFixedSize(24, 24)
        btn_settings.setIcon(self.icon_settings)
        btn_settings.setIconSize(QSize(12, 12))
        btn_settings.setToolTip("Badger settings")

        hbox.addWidget(summary, 1)
        hbox.addWidget(btn_settings)

    def config_logic(self):
        self.btn_settings.clicked.connect(self.go_settings)

    def set_summary(self, text):
        self.summary.setText(text)
        self.setToolTip(text)

    def go_settings(self):
        dlg = BadgerSettingsDialog(self)
        dlg.exec()



import torch
from badger import environment
from badger.errors import BadgerNoInterfaceError


class Environment(environment.Environment):
    name = "multiobjective_test"
    variables = {f"x{i}": [-1, 1] for i in range(20)}
    observables = ["f1", "f2"]

    flag: int = 0

    def set_variables(self, variable_inputs: dict[str, float]):
        if not self.interface:
            raise BadgerNoInterfaceError

        self.interface.set_values(variable_inputs)
        full_outputs = self.interface.get_values(self.variable_names)

        # Filling up the observations
        x = torch.tensor([full_outputs[f"x{i}"] for i in range(20)])
        self.interface.set_values(
            {
                "f1": float((x**2).sum().numpy()),
                "f2": float((x**3).sum().numpy()),
            }
        )



"""
observers...
"""

from datetime import datetime
from badger.logger.event import Events


class Observer:
    def update(self, event, solution):
        raise NotImplementedError


class _Tracker(object):
    def __init__(self):
        self._iterations = 0

        self._start_time = None
        self._previous_time = None

    def _update_tracker(self, event, solution):
        if event == Events.OPTIMIZATION_STEP:
            self._iterations += 1

    def _time_metrics(self):
        now = datetime.now()
        if self._start_time is None:
            self._start_time = now
        if self._previous_time is None:
            self._previous_time = now

        time_elapsed = now - self._start_time
        time_delta = now - self._previous_time

        self._previous_time = now
        return (
            now.strftime("%Y-%m-%d %H:%M:%S"),
            time_elapsed.total_seconds(),
            time_delta.total_seconds(),
        )



from PyQt5.QtWidgets import QLineEdit


def search_bar():
    # completer = QCompleter(word_list)

    line_edit = QLineEdit()
    line_edit.setPlaceholderText("Search")
    # line_edit.setCompleter(completer)

    return line_edit



import multiprocessing as mp
import sys
from unittest.mock import MagicMock, patch

import pytest
from PyQt5.QtWidgets import QApplication

app = QApplication(sys.argv)


@pytest.fixture
def process_creator():
    from badger.gui.default.components.create_process import CreateProcess

    return CreateProcess()


def test_create_subprocess_emits_signals(qtbot, process_creator):
    with (
        patch("badger.gui.default.components.create_process.Process") as mock_process,
        patch("badger.gui.default.components.create_process.run_routine_subprocess"),
    ):
        mock_process.return_value = MagicMock()

        with (
            qtbot.waitSignal(
                process_creator.subprocess_prepared
            ) as blocker_subprocess_prepared,
            qtbot.waitSignal(process_creator.finished) as blocker_finished,
        ):
            process_creator.create_subprocess()

        assert blocker_subprocess_prepared.signal_triggered
        assert blocker_finished.signal_triggered
        mock_process.assert_called_once()

        # Verify that the emitted object contains the expected keys
        emitted_args = blocker_subprocess_prepared.args[0]
        assert set(emitted_args.keys()) == {
            "process",
            "stop_event",
            "pause_event",
            "data_queue",
            "evaluate_queue",
            "wait_event",
        }

        assert isinstance(emitted_args["data_queue"], mp.queues.Queue)
        assert isinstance(emitted_args["evaluate_queue"], tuple)
        assert isinstance(emitted_args["wait_event"], mp.synchronize.Event)
        assert isinstance(emitted_args["stop_event"], mp.synchronize.Event)
        assert isinstance(emitted_args["pause_event"], mp.synchronize.Event)



from typing import Optional
from xopt import VOCS

from PyQt5.QtWidgets import (
    QTableWidgetItem,
)

from badger.routine import Routine

import logging

logger = logging.getLogger(__name__)


class ModelLogic:
    def __init__(self, routine: Optional[Routine], vocs: Optional[VOCS]):
        self.routine = routine
        self.vocs = vocs

    def update_routine(self, routine: Optional[Routine]):
        if routine is not None:
            self.routine = routine
            self.vocs = routine.vocs
        else:
            self.routine = None
            self.vocs = None
            logger.warning("Routine is None")

    def get_reference_points(
        self, ref_inputs: list[QTableWidgetItem], variable_names: list[str]
    ):
        reference_point: dict[str, float] = {}
        if not self.vocs or not ref_inputs:
            return (
                reference_point  # Return empty if vocs or ref_inputs are not available
            )

        # Create a mapping from variable names to ref_inputs
        ref_inputs_dict = dict(zip(self.vocs.variable_names, ref_inputs))
        for var in self.vocs.variable_names:
            if var not in variable_names:
                ref_value = float(ref_inputs_dict[var].text())
                reference_point[var] = ref_value
        return reference_point



# Borrowed from bayes_opt
class Colours:
    """Print in nice colours."""

    BLUE = "\033[94m"
    BOLD = "\033[1m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    END = "\033[0m"
    GREEN = "\033[92m"
    PURPLE = "\033[95m"
    RED = "\033[91m"
    UNDERLINE = "\033[4m"
    YELLOW = "\033[93m"

    @classmethod
    def _wrap_colour(cls, s, colour):
        return colour + s + cls.END

    @classmethod
    def black(cls, s):
        """Wrap text in black."""
        return cls._wrap_colour(s, cls.END)

    @classmethod
    def blue(cls, s):
        """Wrap text in blue."""
        return cls._wrap_colour(s, cls.BLUE)

    @classmethod
    def bold(cls, s):
        """Wrap text in bold."""
        return cls._wrap_colour(s, cls.BOLD)

    @classmethod
    def cyan(cls, s):
        """Wrap text in cyan."""
        return cls._wrap_colour(s, cls.CYAN)

    @classmethod
    def darkcyan(cls, s):
        """Wrap text in darkcyan."""
        return cls._wrap_colour(s, cls.DARKCYAN)

    @classmethod
    def green(cls, s):
        """Wrap text in green."""
        return cls._wrap_colour(s, cls.GREEN)

    @classmethod
    def purple(cls, s):
        """Wrap text in purple."""
        return cls._wrap_colour(s, cls.PURPLE)

    @classmethod
    def red(cls, s):
        """Wrap text in red."""
        return cls._wrap_colour(s, cls.RED)

    @classmethod
    def underline(cls, s):
        """Wrap text in underline."""
        return cls._wrap_colour(s, cls.UNDERLINE)

    @classmethod
    def yellow(cls, s):
        """Wrap text in yellow."""
        return cls._wrap_colour(s, cls.YELLOW)



from copy import deepcopy

from PyQt5.QtWidgets import (
    QDialog,
    QWidget,
    QHBoxLayout,
    QPushButton,
    QVBoxLayout,
    QDoubleSpinBox,
    QRadioButton,
)
from PyQt5.QtWidgets import (
    QGroupBox,
    QLabel,
    QComboBox,
    QStyledItemDelegate,
    QStackedWidget,
)
from PyQt5.QtCore import Qt


class BadgerLimitVariableRangeDialog(QDialog):
    def __init__(
        self, parent, set_vrange, save_config, configs=None, apply_to_all=False
    ):
        super().__init__(parent)

        self.set_vrange = set_vrange
        self.save_config = save_config
        self.configs = deepcopy(configs)
        if configs is None:
            self.configs = {
                "limit_option_idx": 0,
                "ratio_curr": 0.1,
                "ratio_full": 0.1,
                "delta": 0.1,
            }
        self.apply_to_all = apply_to_all

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        self.setWindowTitle("Set variable range")
        self.setMinimumWidth(360)

        vbox = QVBoxLayout(self)

        # Action bar
        action_bar = QWidget()
        hbox = QHBoxLayout(action_bar)
        hbox.setContentsMargins(0, 0, 0, 0)

        lbl = QLabel("Set range by")
        # lbl.setFixedWidth(128)
        self.cb = cb = QComboBox()
        tooltip = """Ratio wrt current value will set the limit
with [ratio * current value] span,
while centering around the current value;

Ratio wrt full range will set the limit
with [ratio * variable range] span,
while centering around the current value;

No matter which option you choose, the limit
will be clipped by the variable range."""
        cb.setToolTip(tooltip)
        cb.setItemDelegate(QStyledItemDelegate())
        cb.addItems(
            [
                "ratio wrt current value",
                "ratio wrt full range",
                "delta around current value",
            ]
        )
        cb.setCurrentIndex(self.configs["limit_option_idx"])

        hbox.addWidget(lbl)
        hbox.addWidget(cb, 1)

        # Config group
        group_config = QGroupBox("Parameters")
        vbox_config = QVBoxLayout(group_config)

        self.stacks = stacks = QStackedWidget()
        # Ratio curr config
        ratio_curr_config = QWidget()
        hbox_ratio_curr = QHBoxLayout(ratio_curr_config)
        hbox_ratio_curr.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel("Ratio")
        self.sb_ratio_curr = sb_ratio_curr = QDoubleSpinBox()
        sb_ratio_curr.setMinimum(0)
        sb_ratio_curr.setMaximum(1)
        sb_ratio_curr.setValue(self.configs["ratio_curr"])
        sb_ratio_curr.setDecimals(2)
        sb_ratio_curr.setSingleStep(0.01)
        hbox_ratio_curr.addWidget(lbl)
        hbox_ratio_curr.addWidget(sb_ratio_curr, 1)
        # Ratio full config
        ratio_full_config = QWidget()
        hbox_ratio_full = QHBoxLayout(ratio_full_config)
        hbox_ratio_full.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel("Ratio")
        self.sb_ratio_full = sb_ratio_full = QDoubleSpinBox()
        sb_ratio_full.setMinimum(0)
        sb_ratio_full.setMaximum(1)
        sb_ratio_full.setValue(self.configs["ratio_full"])
        sb_ratio_full.setDecimals(2)
        sb_ratio_full.setSingleStep(0.01)
        hbox_ratio_full.addWidget(lbl)
        hbox_ratio_full.addWidget(sb_ratio_full, 1)
        # Delta config
        delta_config = QWidget()
        hbox_delta = QHBoxLayout(delta_config)
        hbox_delta.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel("Delta")
        self.sb_delta = sb_delta = QDoubleSpinBox()
        sb_delta.setMinimum(0)
        sb_delta.setMaximum(1e4)
        sb_delta.setValue(self.configs["delta"])
        sb_delta.setDecimals(4)
        sb_delta.setSingleStep(0.01)
        hbox_delta.addWidget(lbl)
        hbox_delta.addWidget(sb_delta, 1)

        stacks.addWidget(ratio_curr_config)
        stacks.addWidget(ratio_full_config)
        stacks.addWidget(delta_config)

        stacks.setCurrentIndex(self.configs["limit_option_idx"])
        vbox_config.addWidget(stacks)
        vbox_config.addStretch(1)

        # Apply to group
        apply_to_group = QWidget()
        hbox_apply_to = QHBoxLayout(apply_to_group)
        hbox_apply_to.setContentsMargins(0, 0, 0, 0)
        lbl_apply_to = QLabel("Apply to:")
        self.rb_all_variables = rb_all_variables = QRadioButton("All Variables")
        self.rb_only_visible = rb_only_visible = QRadioButton("Only Visible")

        if self.apply_to_all:
            rb_all_variables.setChecked(True)
        else:
            rb_only_visible.setChecked(True)  # Default selection
        hbox_apply_to.addWidget(lbl_apply_to)
        hbox_apply_to.addWidget(rb_all_variables)
        hbox_apply_to.addWidget(rb_only_visible)
        vbox_config.addWidget(apply_to_group)

        # Button set
        button_set = QWidget()
        hbox_set = QHBoxLayout(button_set)
        hbox_set.setContentsMargins(0, 0, 0, 0)
        self.btn_cancel = btn_cancel = QPushButton("Cancel")
        self.btn_set = btn_set = QPushButton("Set")
        btn_cancel.setFixedSize(96, 24)
        btn_set.setFixedSize(96, 24)
        hbox_set.addStretch()
        hbox_set.addWidget(btn_cancel)
        hbox_set.addWidget(btn_set)

        vbox.addWidget(action_bar)
        vbox.addWidget(group_config, 1)
        vbox.addWidget(button_set)

    def config_logic(self):
        self.cb.currentIndexChanged.connect(self.limit_option_changed)
        self.btn_cancel.clicked.connect(self.close)
        self.btn_set.clicked.connect(
            lambda: self.set(set_all=self.rb_all_variables.isChecked())
        )
        self.sb_ratio_curr.valueChanged.connect(self.ratio_curr_changed)
        self.sb_ratio_full.valueChanged.connect(self.ratio_full_changed)
        self.sb_delta.valueChanged.connect(self.delta_changed)

    def ratio_curr_changed(self, ratio_curr):
        self.configs["ratio_curr"] = ratio_curr

    def ratio_full_changed(self, ratio_full):
        self.configs["ratio_full"] = ratio_full

    def delta_changed(self, delta):
        self.configs["delta"] = delta

    def set(self, set_all: bool = True):
        self.save_config(self.configs)
        self.set_vrange(set_all=set_all)  # pass set_all flag to set_vrange
        self.close()

    def limit_option_changed(self, i):
        self.stacks.setCurrentIndex(i)

        # Update configs
        self.configs["limit_option_idx"] = i

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            self.set(set_all=self.rb_all_variables.isChecked())
        elif event.key() == Qt.Key_Escape:
            self.close()



import os
from importlib import metadata
from typing import Dict

from PyQt5.QtCore import QThread
from PyQt5.QtWidgets import QDesktopWidget, QMainWindow, QMessageBox, QStackedWidget

from badger.gui.default.components.create_process import CreateProcess
from badger.gui.default.components.process_manager import ProcessManager
from badger.gui.default.pages.home_page import BadgerHomePage


class BadgerMainWindow(QMainWindow):
    def __init__(self) -> None:
        super().__init__()
        self.thread_list = []
        self.process_manager = ProcessManager()
        self.process_manager.processQueueUpdated.connect(self.addSubprocess)
        self.addSubprocess()
        self.init_ui()
        self.config_logic()

    def addSubprocess(self) -> None:
        """
        Adds a subprocess to the subprocess queue.
        This method builds the subprocess on a QThread so as to not disrupt the main process.
        """
        self.thread = QThread()
        self.worker = CreateProcess()
        self.worker.moveToThread(self.thread)

        self.thread.started.connect(self.worker.create_subprocess)
        self.worker.subprocess_prepared.connect(self.storeSubprocess)
        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)
        self.thread.finished.connect(self.cleanupThread)

        self.thread_list.append(self.thread)
        self.thread.start()

    def cleanupThread(self) -> None:
        """
        Method to remove threads no longer active from the thread list.

        Parameters:
            thread: QThread
        """
        thread = self.sender()
        if thread in self.thread_list:
            self.thread_list.remove(thread)

    def storeSubprocess(self, process_with_args: Dict) -> None:
        """
        Store the prepared subprocess for later use.

        Parameters:
            process_with_args: Dict
        """
        self.process_manager.add_to_queue(process_with_args)

    def init_ui(self) -> None:
        version = metadata.version("badger-opt")
        version_xopt = metadata.version("xopt")
        self.setWindowTitle(f"Badger v{version} (Xopt v{version_xopt})")
        if os.getenv("DEMO"):
            self.resize(1280, 720)
        else:
            self.resize(1080, 720)
        self.center()

        # Add menu bar
        # menu_bar = self.menuBar()
        # edit_menu = menu_bar.addMenu('Edit')
        # edit_menu.addAction('New')

        # Add pages
        self.home_page = BadgerHomePage(self.process_manager)

        self.stacks = stacks = QStackedWidget()
        stacks.addWidget(self.home_page)

        stacks.setCurrentIndex(0)

        self.setCentralWidget(self.stacks)

    def center(self) -> None:
        qr = self.frameGeometry()
        cp = QDesktopWidget().availableGeometry().center()

        qr.moveCenter(cp)
        self.move(qr.topLeft())

    def config_logic(self) -> None:
        pass

    def closeEvent(self, event) -> None:
        monitor = self.home_page.run_monitor
        if not monitor.running:
            self.process_manager.close_proccesses()
            monitor.destroy_unused_env()
            return

        reply = QMessageBox.question(
            self,
            "Window Close",
            "Closing this window will terminate the current run, "
            "and the run data would be archived.\n\nProceed?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )

        if reply == QMessageBox.Yes:

            def close_window():
                monitor.destroy_unused_env()
                self.close()

            monitor.register_post_run_action(close_window)
            monitor.testing = True  # suppress the archive pop-ups
            monitor.routine_runner.stop_routine()
            event.ignore()
        else:
            event.ignore()



# from PyQt5.QtCore import QRegExp
# from PyQt5.QtGui import QRegExpValidator
from PyQt5.QtWidgets import (
    # QComboBox,
    QGridLayout,
    QVBoxLayout,
    QWidget,
    QLabel,
    QLineEdit,
)
from PyQt5.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QApplication,
)
from qdarkstyle import load_stylesheet, DarkPalette, LightPalette
from badger.settings import init_settings


class BadgerSettingsDialog(QDialog):
    theme_list = ["default", "light", "dark"]
    theme_idx_dict = {
        "default": 0,
        "light": 1,
        "dark": 2,
    }

    def __init__(self, parent):
        super().__init__(parent)

        self.config_singleton = init_settings()
        self.settings = (
            self.config_singleton.list_settings()
        )  # store the current settings

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        self.setWindowTitle("Badger settings")
        self.setMinimumWidth(480)

        vbox = QVBoxLayout(self)

        # validator = QRegExpValidator(QRegExp(r"^[0-9]\d*(\.\d+)?$"))

        widget_settings = QWidget(self)
        grid = QGridLayout(widget_settings)
        grid.setContentsMargins(0, 0, 0, 0)

        # Theme selector
        # theme = self.settings["BADGER_THEME"]["value"]
        # self.lbl_theme = lbl_theme = QLabel("Theme")
        # self.cb_theme = cb_theme = QComboBox()
        # cb_theme.setItemDelegate(QStyledItemDelegate())
        # cb_theme.addItems(self.theme_list)
        # cb_theme.setCurrentIndex(self.theme_idx_dict[theme])
        # grid.addWidget(lbl_theme, 0, 0)
        # grid.addWidget(cb_theme, 0, 1)

        # Plugin Root
        self.plugin_root = plugin_root = QLabel("Plugin Root")
        self.plugin_root_path = plugin_root_path = QLineEdit(
            self.config_singleton.read_value("BADGER_PLUGIN_ROOT")
        )
        grid.addWidget(plugin_root, 1, 0)
        grid.addWidget(plugin_root_path, 1, 1)

        # Template Root
        self.template_root = template_root = QLabel("Template Root")
        self.template_root_path = template_root_path = QLineEdit(
            self.config_singleton.read_value("BADGER_TEMPLATE_ROOT")
        )
        grid.addWidget(template_root, 2, 0)
        grid.addWidget(template_root_path, 2, 1)

        # Logbook Root
        self.logbook_root = logbook_root = QLabel("Logbook Root")
        self.logbook_root_path = logbook_root_path = QLineEdit(
            self.config_singleton.read_value("BADGER_LOGBOOK_ROOT")
        )
        grid.addWidget(logbook_root, 3, 0)
        grid.addWidget(logbook_root_path, 3, 1)

        # Archive Root
        self.archive_root = archive_root = QLabel("Archive Root")
        self.archive_root_path = archive_root_path = QLineEdit(
            self.config_singleton.read_value("BADGER_ARCHIVE_ROOT")
        )
        grid.addWidget(archive_root, 4, 0)
        grid.addWidget(archive_root_path, 4, 1)

        # Auto refresh
        # self.auto_refresh = auto_refresh = QLabel("Auto Refresh")
        # self.enable_auto_refresh = enable_auto_refresh = QCheckBox()
        # enable_auto_refresh.setChecked(
        #     strtobool(self.config_singleton.read_value("AUTO_REFRESH"))
        # )
        # grid.addWidget(auto_refresh, 5, 0)
        # grid.addWidget(enable_auto_refresh, 5, 1)

        # Check Variable Interval
        # self.var_int = var_int = QLabel('Check Variable Interval')
        # self.var_int_val = var_int_val = QLineEdit(str(read_value('BADGER_CHECK_VAR_INTERVAL')))
        # self.var_int_val.setValidator(validator)
        # grid.addWidget(var_int, 5, 0)
        # grid.addWidget(var_int_val, 5, 1)

        # Check Variable Timeout
        # self.var_time = var_time = QLabel('Check Variable Timeout')
        # self.var_time_val = var_time_val = QLineEdit(str(read_value('BADGER_CHECK_VAR_TIMEOUT')))
        # self.var_time_val.setValidator(validator)
        # grid.addWidget(var_time, 6, 0)
        # grid.addWidget(var_time_val, 6, 1)

        # Plugin URL
        # self.plugin_url = plugin_url = QLabel('Plugin Server URL')
        # self.plugin_url_name = plugin_url_name = QLineEdit(read_value('BADGER_PLUGINS_URL'))
        # grid.addWidget(plugin_url, 7, 0)
        # grid.addWidget(plugin_url_name, 7, 1)

        # Badger data dump period
        # self.dump_period = dump_period = QLabel("Data dump period")
        # self.dump_period_val = dump_period_val = QLineEdit(
        #     str(self.config_singleton.read_value("BADGER_DATA_DUMP_PERIOD"))
        # )
        # self.dump_period_val.setValidator(validator)
        # grid.addWidget(dump_period, 8, 0)
        # grid.addWidget(dump_period_val, 8, 1)

        # Advanced settings
        # self.adv_features = adv_features = QLabel("Enable Advanced Features")
        # self.enable_adv_features = enable_adv_features = QCheckBox()
        # enable_adv_features.setChecked(
        #     strtobool(self.config_singleton.read_value("BADGER_ENABLE_ADVANCED"))
        # )
        # grid.addWidget(adv_features, 9, 0)
        # grid.addWidget(enable_adv_features, 9, 1)

        grid.setColumnStretch(1, 1)

        vbox.addWidget(widget_settings)

        self.btns = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)  # type: ignore

        vbox.addStretch(1)
        vbox.addWidget(self.btns)

    def config_logic(self):
        # self.cb_theme.currentIndexChanged.connect(self.select_theme)
        self.btns.accepted.connect(self.apply_settings)
        self.btns.rejected.connect(self.restore_settings)

    def set_theme(self, theme):
        app = QApplication.instance()
        if theme == "dark":
            app.setStyleSheet(load_stylesheet(palette=DarkPalette))
        elif theme == "light":
            app.setStyleSheet(load_stylesheet(palette=LightPalette))
        else:
            app.setStyleSheet("")

    def select_theme(self, i):
        theme = self.theme_list[i]
        self.set_theme(theme)
        # Update the internal settings
        self.config_singleton.write_value("BADGER_THEME", theme)

    def apply_settings(self):
        self.accept()
        self.config_singleton.write_value(
            "BADGER_PLUGIN_ROOT", self.plugin_root_path.text()
        )
        self.config_singleton.write_value(
            "BADGER_TEMPLATE_ROOT", self.template_root_path.text()
        )
        self.config_singleton.write_value(
            "BADGER_LOGBOOK_ROOT", self.logbook_root_path.text()
        )
        self.config_singleton.write_value(
            "BADGER_ARCHIVE_ROOT", self.archive_root_path.text()
        )
        # self.config_singleton.write_value(
        #     "AUTO_REFRESH", self.enable_auto_refresh.isChecked()
        # )
        # write_value('BADGER_CHECK_VAR_INTERVAL', self.var_int_val.text())
        # write_value('BADGER_CHECK_VAR_TIMEOUT', self.var_time_val.text())
        # write_value('BADGER_PLUGINS_URL', self.plugin_url_name.text())
        # self.config_singleton.write_value(
        #     "BADGER_DATA_DUMP_PERIOD", float(self.dump_period_val.text())
        # )
        # self.config_singleton.write_value(
        #     "BADGER_ENABLE_ADVANCED", self.enable_adv_features.isChecked()
        # )

    def restore_settings(self):
        # Reset theme if needed
        theme_curr = self.config_singleton.read_value("BADGER_THEME")
        theme_prev = self.settings["BADGER_THEME"]["value"]
        if theme_prev != theme_curr:
            self.set_theme(theme_prev)

        for key in self.settings.keys():
            self.config_singleton.write_value(key, self.settings[key]["value"])

        self.reject()



from PyQt5.QtWidgets import QDialog, QPlainTextEdit, QVBoxLayout, QWidget
from PyQt5.QtWidgets import QHBoxLayout, QPushButton
from PyQt5.QtGui import QFont
from badger.gui.default.components.syntax import PythonHighlighter


class BadgerEditScriptDialog(QDialog):
    def __init__(self, parent, generator, script, callback):
        super().__init__(parent)

        self.generator = generator
        self.script = script
        self.callback = callback

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        self.setWindowTitle(f"Edit Script for {self.generator}")
        self.resize(640, 640)

        vbox = QVBoxLayout(self)

        self.script_editor = script_editor = QPlainTextEdit()
        self.highlighter = PythonHighlighter(script_editor.document())
        font = QFont("Menlo", 13)
        font.setFixedPitch(True)
        script_editor.setFont(font)
        script_editor.setPlainText(self.script)

        # Button set
        button_set = QWidget()
        hbox_set = QHBoxLayout(button_set)
        hbox_set.setContentsMargins(0, 0, 0, 0)
        self.btn_cancel = btn_cancel = QPushButton("Cancel")
        self.btn_save = btn_save = QPushButton("Save")
        btn_save.setDisabled(True)
        btn_cancel.setFixedSize(96, 24)
        btn_save.setFixedSize(96, 24)
        hbox_set.addStretch()
        hbox_set.addWidget(btn_cancel)
        hbox_set.addWidget(btn_save)

        vbox.addWidget(script_editor)
        vbox.addWidget(button_set)

    def config_logic(self):
        self.script_editor.textChanged.connect(self.text_changed)
        self.btn_cancel.clicked.connect(self.cancel_changes)
        self.btn_save.clicked.connect(self.save_changes)

    def text_changed(self):
        self.btn_save.setDisabled(False)

    def cancel_changes(self):
        self.callback(self.script)
        self.accept()

    def save_changes(self):
        self.callback(self.script_editor.toPlainText())
        self.accept()



from PyQt5.QtWidgets import QMessageBox
import traceback
import sys


class BadgerError(Exception):
    def __init__(self, message="", detailed_text=None):
        if detailed_text is None:
            detailed_text = self.capture_traceback_or_stack()

        super().__init__(message)
        self.detailed_text = detailed_text
        self.show_message_box()

    def show_message_box(self):
        """
        Method to create and display a popup window with the error message.
        """
        from badger.gui.default.windows.expandable_message_box import (
            ExpandableMessageBox,
        )

        error_message = str(self)
        dialog = ExpandableMessageBox(
            text=error_message, detailedText=self.detailed_text
        )
        dialog.setIcon(QMessageBox.Critical)
        dialog.exec_()

    def capture_traceback_or_stack(self):
        """
        Captures the current traceback if an exception is active, otherwise captures the call stack.
        """
        exc_type, exc_value, exc_traceback = sys.exc_info()
        if exc_traceback:
            return "".join(
                traceback.format_exception(exc_type, exc_value, exc_traceback)
            )
        else:
            return "".join(traceback.format_stack())


class BadgerConfigError(Exception):
    pass


class VariableRangeError(Exception):
    pass


class BadgerNotImplementedError(Exception):
    pass


class BadgerDBError(Exception):
    pass


class BadgerEnvInstantiationError(Exception):
    pass


class BadgerEnvVarError(Exception):
    pass


class BadgerEnvObsError(Exception):
    pass


class BadgerNoInterfaceError(Exception):
    def __init__(self, message="Must provide an interface!"):
        super().__init__(message)


class BadgerInterfaceChannelError(Exception):
    pass


class BadgerInvalidPluginError(Exception):
    pass


class BadgerPluginNotFoundError(Exception):
    pass


class BadgerInvalidDocsError(Exception):
    pass


class BadgerLogbookError(Exception):
    pass


class BadgerLoadConfigError(Exception):
    def __init__(self, message="Config Error!"):
        super().__init__(message)


class BadgerRoutineError(Exception):
    pass


class BadgerRunTerminated(Exception):
    def __init__(self, message="Optimization run has been terminated!"):
        super().__init__(message)



from PyQt5.QtWidgets import (
    QDialog,
    QMessageBox,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QTextEdit,
)
from PyQt5.QtGui import QTextOption, QFont, QFontDatabase


class ExpandableMessageBox(QDialog):
    def __init__(
        self, icon=None, title="Message", text="", detailedText="", parent=None
    ):
        super().__init__(parent)

        # Main layout
        mainLayout = QVBoxLayout(self)

        # Top layout for icon and main text
        topLayout = QHBoxLayout()
        self.iconLabel = QLabel()
        if icon:
            self.iconLabel.setPixmap(icon.pixmap(64, 64))
        self.textLabel = QLabel(text)
        self.textLabel.setMinimumWidth(280)
        self.textLabel.setWordWrap(True)
        font = QFont()
        font.setBold(True)
        self.textLabel.setFont(font)
        topLayout.addWidget(self.iconLabel)
        topLayout.addWidget(self.textLabel, 1)
        mainLayout.addLayout(topLayout)

        # Detailed text area setup
        self.detailedTextWidget = QTextEdit()
        self.detailedTextWidget.setText(detailedText)
        self.detailedTextWidget.setReadOnly(True)
        self.detailedTextWidget.setWordWrapMode(
            QTextOption.WrapAtWordBoundaryOrAnywhere
        )
        monoFont = QFontDatabase.systemFont(QFontDatabase.FixedFont)
        monoFont.setPointSize(9)
        self.detailedTextWidget.setFont(monoFont)
        self.detailedTextWidget.setVisible(False)  # Initially hidden

        # Button to show/hide details
        self.toggleButton = QPushButton("Show Details")
        self.toggleButton.clicked.connect(self.toggle_details)

        # Layouts for the detailed text and button
        self.detailLayout = QVBoxLayout()
        self.detailLayout.addWidget(self.detailedTextWidget)
        self.detailLayout.addWidget(self.toggleButton)

        # Add the detailed text area and toggle button to the main layout
        mainLayout.addLayout(self.detailLayout)

        # Buttons
        self.buttonBox = QHBoxLayout()
        self.okButton = QPushButton("OK")
        self.okButton.clicked.connect(self.accept)
        self.buttonBox.addWidget(self.okButton)
        mainLayout.addLayout(self.buttonBox)

        # Set window properties
        self.setWindowTitle(title)
        self.resize(420, 250)

    def toggle_details(self):
        if self.detailedTextWidget.isVisible():
            self.detailedTextWidget.setVisible(False)
            self.toggleButton.setText("Show Details")
        else:
            self.detailedTextWidget.setVisible(True)
            self.toggleButton.setText("Hide Details")

    def setText(self, text):
        self.textLabel.setText(text)

    def setDetailedText(self, detailedText):
        self.detailedTextWidget.setText(detailedText)

    def setIcon(self, icon):
        # This maps the QMessageBox icons to the QDialog
        iconMap = {
            QMessageBox.Information: QMessageBox.standardIcon(QMessageBox.Information),
            QMessageBox.Warning: QMessageBox.standardIcon(QMessageBox.Warning),
            QMessageBox.Critical: QMessageBox.standardIcon(QMessageBox.Critical),
            QMessageBox.Question: QMessageBox.standardIcon(QMessageBox.Question),
        }
        standardIcon = iconMap.get(icon)
        if standardIcon:
            self.iconLabel.setPixmap(standardIcon)



import logging

from badger.utils import yprint

logger = logging.getLogger(__name__)


def show_generator(args):
    try:
        from badger.factory import list_generators, get_generator
    except Exception as e:
        logger.error(e)
        return

    if args.generator_name is None:
        yprint(list_generators())
        return

    try:
        configs = get_generator(args.generator_name)
        yprint(configs)
    except Exception as e:
        logger.error(e)
        try:
            # The exception could carry the configs information
            configs = e.configs
            yprint(configs)
        except:
            pass



import pytest


@pytest.fixture
def process_manager():
    from badger.gui.default.components.process_manager import ProcessManager

    return ProcessManager()


class TestProcessManager:
    def test_add_to_queue(self, process_manager):
        """
        Test that a process can be added to the queue.
        """
        process_with_args = {"process": "test_process", "args": [1, 2, 3]}
        process_manager.add_to_queue(process_with_args)
        assert len(process_manager.processes_queue) == 1
        assert process_manager.processes_queue[0] == process_with_args

    def test_remove_from_queue(self, process_manager, qtbot):
        """
        Test that a process can be removed from the queue and the correct process is returned.
        """
        process_with_args1 = {"process": "test_process_one", "args": [1, 2, 3]}
        process_with_args2 = {"process": "test_process_two", "args": [4, 5, 6]}
        process_manager.add_to_queue(process_with_args1)
        process_manager.add_to_queue(process_with_args2)

        removed_process = process_manager.remove_from_queue()
        assert removed_process == process_with_args1
        assert len(process_manager.processes_queue) == 1
        removed_process = process_manager.remove_from_queue()
        assert removed_process == process_with_args2
        removed_process = process_manager.remove_from_queue()
        assert removed_process is None
        assert len(process_manager.processes_queue) == 0

        process_manager.add_to_queue(process_with_args1)

        with qtbot.waitSignal(process_manager.processQueueUpdated) as blocker:
            process_manager.remove_from_queue()

        assert blocker.signal_triggered



from PyQt5.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QComboBox,
    QLabel,
    QGroupBox,
    QTableWidget,
    QTableWidgetItem,
    QSpinBox,
    QPushButton,
    QCheckBox,
    QHeaderView,
)
from PyQt5.QtCore import Qt
from typing import Optional, cast

from badger.gui.default.components.bo_visualizer.types import ConfigurableOptions


from xopt import VOCS

import logging


logger = logging.getLogger(__name__)


class UIComponents:
    def __init__(
        self,
        default_parameters: ConfigurableOptions,
        vocs: Optional[VOCS] = None,
    ):
        self.vocs = vocs
        self.variable_checkboxes = {}
        self.ref_inputs: list[QTableWidgetItem] = []
        self.reference_table = None  # Will be initialized later

        # Initialize other UI components
        self.update_button = QPushButton("Update")
        self.x_axis_combo = QComboBox()
        self.y_axis_combo = QComboBox()
        self.y_axis_checkbox = QCheckBox("Include Variable 2")
        self.acq_func_checkbox = QCheckBox("Show Acquisition Function")
        self.show_samples_checkbox = QCheckBox("Show Samples")
        self.show_prior_mean_checkbox = QCheckBox("Show Prior Mean")
        self.show_feasibility_checkbox = QCheckBox("Show Feasibility")
        self.n_grid = QSpinBox()

        include_variable_2 = default_parameters["include_variable_2"]
        show_prior_mean = default_parameters["plot_options"]["show_prior_mean"]
        show_feasibility = default_parameters["plot_options"]["show_feasibility"]
        show_samples = default_parameters["plot_options"]["show_samples"]
        show_acq_func = default_parameters["plot_options"]["show_acq_func"]
        n_grid = default_parameters["plot_options"]["n_grid"]
        n_grid_range = default_parameters["plot_options"]["n_grid_range"]

        # Set default parameters
        self.y_axis_checkbox.setChecked(include_variable_2)
        self.acq_func_checkbox.setChecked(show_acq_func)
        self.show_samples_checkbox.setChecked(show_samples)
        self.show_prior_mean_checkbox.setChecked(show_prior_mean)
        self.show_feasibility_checkbox.setChecked(show_feasibility)
        self.n_grid.setRange(n_grid_range[0], n_grid_range[1])
        self.n_grid.setValue(n_grid)

        self.ref_inputs = []

        # Initialize layouts
        self.variable_checkboxes_layout = None

    def create_variable_checkboxes(self):
        group_box = QGroupBox("Select Variables")
        layout = self.variable_checkboxes_layout or QVBoxLayout()
        group_box.setLayout(layout)
        return group_box

    def create_axis_layout(self):
        layout = QVBoxLayout()

        x_layout = QHBoxLayout()
        x_label = QLabel("Variable 1:")
        x_axis_label = QLabel("X-axis")
        x_layout.addWidget(x_label)
        x_layout.addWidget(self.x_axis_combo)
        x_layout.addWidget(x_axis_label)

        y_layout = QHBoxLayout()
        y_label = QLabel("Variable 2:")
        y_axis_label = QLabel("Y-axis")
        y_layout.addWidget(y_label)
        y_layout.addWidget(self.y_axis_combo)
        y_layout.addWidget(y_axis_label)

        include_var2_layout = QHBoxLayout()
        include_var2_layout.addWidget(self.y_axis_checkbox)

        layout.addLayout(x_layout)
        layout.addLayout(y_layout)
        layout.addLayout(include_var2_layout)

        return layout

    def initialize_ui_components(self, configurable_options: ConfigurableOptions):
        self.populate_reference_table()

    def create_reference_inputs(self):
        group_box = QGroupBox("Reference Points")
        layout = QVBoxLayout()

        self.reference_table = QTableWidget()
        self.reference_table.setColumnCount(2)
        self.reference_table.setHorizontalHeaderLabels(["Variable", "Ref. Point"])
        horizontal_header = self.reference_table.horizontalHeader()
        if horizontal_header is not None:
            horizontal_header.setSectionResizeMode(QHeaderView.ResizeMode.Stretch)

        layout.addWidget(self.reference_table)
        group_box.setLayout(layout)
        return group_box

    def populate_reference_table(self):
        """Populate the reference table based on the current vocs variable names."""

        logger.debug("Populating reference table")
        if self.reference_table is None:
            raise Exception("Reference Table is None")

        if self.vocs is None:
            raise Exception("Vocs is None")

        self.reference_table.blockSignals(True)

        self.reference_table.setRowCount(len(self.vocs.variable_names))
        self.ref_inputs = []

        for i, var_name in enumerate(self.vocs.variable_names):
            variable_item = QTableWidgetItem(var_name)
            itemIsEditable = Qt.ItemFlag.ItemIsEditable

            variable_item.setFlags(
                variable_item.flags() & ~Qt.ItemFlags(itemIsEditable)
            )
            self.reference_table.setItem(i, 0, variable_item)

            # Set default reference point to the midpoint of variable bounds
            default_value = cast(
                float,
                (self.vocs.variables[var_name][0] + self.vocs.variables[var_name][1])
                / 2,
            )
            reference_point_item = QTableWidgetItem(str(default_value))
            self.ref_inputs.append(reference_point_item)
            self.reference_table.setItem(i, 1, reference_point_item)
        self.reference_table.blockSignals(False)

    def create_options_section(self):
        group_box = QGroupBox("Plot Options")
        layout = QVBoxLayout()

        layout.addWidget(self.acq_func_checkbox)
        layout.addWidget(self.show_samples_checkbox)
        layout.addWidget(self.show_prior_mean_checkbox)
        layout.addWidget(self.show_feasibility_checkbox)

        grid_layout = QHBoxLayout()
        n_grid_label = QLabel("No. of Grid Points:")
        grid_layout.addWidget(n_grid_label)
        grid_layout.addWidget(self.n_grid)

        layout.addLayout(grid_layout)
        group_box.setLayout(layout)
        return group_box

    def create_buttons(self):
        layout = QHBoxLayout()
        self.update_button = QPushButton("Update")

        layout.addWidget(self.update_button)

        self.update_button.setObjectName("update_button")

        return layout

    def update_variables(self, configurable_options: ConfigurableOptions):
        if self.vocs is None:
            raise Exception("Vocs is None")

        x_axis_items = [
            self.x_axis_combo.itemText(i) for i in range(self.x_axis_combo.count())
        ]

        y_axis_items = [
            self.y_axis_combo.itemText(i) for i in range(self.y_axis_combo.count())
        ]

        combined_set = set(x_axis_items + y_axis_items)

        if self.vocs.variable_names != list(combined_set):
            logger.debug(
                f"Populating axis combos with variable names: {self.vocs.variable_names}"
            )

            self.x_axis_combo.blockSignals(True)
            self.y_axis_combo.blockSignals(True)

            self.x_axis_combo.clear()
            self.y_axis_combo.clear()

            self.x_axis_combo.addItems(self.vocs.variable_names)
            self.y_axis_combo.addItems(self.vocs.variable_names)

            self.x_axis_combo.setCurrentIndex(configurable_options["variable_1"])
            self.y_axis_combo.setCurrentIndex(configurable_options["variable_2"])

            self.x_axis_combo.blockSignals(False)
            self.y_axis_combo.blockSignals(False)

            self.populate_reference_table()

    def update_vocs(
        self,
        vocs: Optional[VOCS],
    ):
        if vocs is None:
            raise Exception("Vocs in None")

        self.vocs = vocs
        logger.debug(f"vocs: {vocs}")



from PyQt5.QtWidgets import QWidget, QVBoxLayout, QTreeWidget, QTreeWidgetItem
from PyQt5.QtGui import QFont
from PyQt5.QtCore import Qt
from badger.archive import get_base_run_filename
from badger.utils import run_names_to_dict


class HistoryNavigator(QWidget):
    def __init__(self):
        super().__init__()

        # Layout for the widget
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self.tree_widget = QTreeWidget()

        self.tree_widget.setHeaderLabels(["History Navigator"])
        header = self.tree_widget.header()
        # Set the font of the header to bold
        bold_font = QFont()
        bold_font.setBold(True)
        header.setFont(bold_font)

        self.tree_widget.setMinimumHeight(256)
        layout.addWidget(self.tree_widget)

        self.runs = None  # all runs to be shown in the tree widget
        self.setStyleSheet("""
            QTreeWidget {
                background-color: #37414F;
            }
        """)

    def _firstSelectableItem(self, parent=None):
        """
        Internal recursive function for finding the first selectable item.
        """
        if parent is None:
            parent = self.tree_widget.invisibleRootItem()

        for i in range(parent.childCount()):
            item = parent.child(i)
            if item.flags() & Qt.ItemIsSelectable:
                return item
            result = self._firstSelectableItem(item)
            if result:
                return result
        return None

    def updateItems(self, runs=None):
        self.tree_widget.clear()
        self.runs = runs  # store the runs for navigation
        if runs is None:
            return

        runs_dict = run_names_to_dict(runs)
        first_items = []
        flag_first_item = True

        for year, dict_year in runs_dict.items():
            item_year = QTreeWidgetItem([year])
            item_year.setFlags(item_year.flags() & ~Qt.ItemIsSelectable)

            if flag_first_item:
                first_items.append(item_year)

            for month, dict_month in dict_year.items():
                item_month = QTreeWidgetItem([month])
                item_month.setFlags(item_month.flags() & ~Qt.ItemIsSelectable)

                if flag_first_item:
                    first_items.append(item_month)

                for day, list_day in dict_month.items():
                    item_day = QTreeWidgetItem([day])
                    item_day.setFlags(item_day.flags() & ~Qt.ItemIsSelectable)

                    if flag_first_item:
                        first_items.append(item_day)
                        flag_first_item = False

                    for file in list_day:
                        item_file = QTreeWidgetItem([file])
                        item_day.addChild(item_file)
                    item_month.addChild(item_day)
                item_year.addChild(item_month)
            self.tree_widget.addTopLevelItem(item_year)

        # Expand the first set of items
        for item in first_items:
            item.setExpanded(True)

    def selectNextItem(self):
        run_curr = get_base_run_filename(self.currentText())
        idx = self.runs.index(run_curr)
        if idx < len(self.runs) - 1:
            self._selectItemByRun(self.runs[idx + 1])

    def selectPreviousItem(self):
        run_curr = get_base_run_filename(self.currentText())
        idx = self.runs.index(run_curr)
        if idx > 0:
            self._selectItemByRun(self.runs[idx - 1])

    def _selectItemByRun(self, run):
        """
        Internal function to select a tree widget item by run name.
        """
        for i in range(self.tree_widget.topLevelItemCount()):
            year_item = self.tree_widget.topLevelItem(i)
            for j in range(year_item.childCount()):
                month_item = year_item.child(j)
                for k in range(month_item.childCount()):
                    day_item = month_item.child(k)
                    for _l in range(day_item.childCount()):
                        file_item = day_item.child(_l)
                        if get_base_run_filename(file_item.text(0)) == run:
                            self.tree_widget.setCurrentItem(file_item)
                            return

    def currentText(self):
        current_item = self.tree_widget.currentItem()
        if current_item:
            return current_item.text(0)
        return ""

    def count(self):
        if self.runs is None:
            return 0

        return len(self.runs)



from importlib import resources
from typing import Any, List, Tuple
from PyQt5.QtCore import (
    pyqtSlot,
    Qt,
)
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (
    QAbstractSpinBox,
    QDoubleSpinBox,
    QTableWidget,
    QHeaderView,
    QAbstractItemView,
    QCheckBox,
    QPushButton,
    QStyledItemDelegate,
    QWidget,
    QHBoxLayout,
)

from badger.gui.default.utils import (
    MouseWheelWidgetAdjustmentGuard,
    NoHoverFocusComboBox,
)


CONS_RELATION_DICT = {
    ">": "GREATER_THAN",
    "<": "LESS_THAN",
}


class ConstraintTable(QTableWidget):
    """
    A custom QTableWidget for displaying and managing constraints.

    This table supports:
      - Displaying constraint names, relations, and thresholds.
      - Toggling constraint criticality via checkboxes.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the ConstraintTable widget.

        Parameters
        ----------
        *args : Any
            Variable length argument list for QTableWidget.
        **kwargs : Any
            Arbitrary keyword arguments for QTableWidget.
        """
        super().__init__(*args, **kwargs)

        icon_ref = resources.files(__package__) / "../images/trash.png"
        with resources.as_file(icon_ref) as icon_path:
            self.icon_trash = QIcon(str(icon_path))

        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setShowGrid(False)

        self.setRowCount(0)
        self.setColumnCount(5)
        self.setAlternatingRowColors(True)
        self.setStyleSheet("alternate-background-color: #262E38;")
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)

        self.setColumnWidth(0, 192)
        self.setColumnWidth(1, 64)
        self.setColumnWidth(3, 64)
        self.setColumnWidth(4, 32)

        self.verticalHeader().setVisible(False)
        self.update_header_visibility()

    def update_header_visibility(self):
        self.setHorizontalHeaderLabels(
            ["Name", "Relation", "Threshold", "Critical", ""]
        )
        self.horizontalHeader().setSectionResizeMode(QHeaderView.Interactive)
        self.horizontalHeader().setSectionResizeMode(0, QHeaderView.Stretch)
        self.horizontalHeader().setVisible(self.rowCount() > 0)

    def add_constraint(
        self, options, name=None, relation=0, threshold=0, critical=False, decimals=4
    ) -> None:
        """
        Adds a constraint.
        """
        currentRow = self.rowCount()
        self.setRowCount(self.rowCount() + 1)

        check_crit = QCheckBox()
        check_crit.setChecked(critical)
        self.setCellWidget(currentRow, 3, check_crit)

        cb_obs = NoHoverFocusComboBox()
        cb_obs.setItemDelegate(QStyledItemDelegate())
        cb_obs.addItems(options)
        try:
            idx = options.index(name)
        except:
            idx = 0
        cb_obs.setCurrentIndex(idx)
        self.setCellWidget(currentRow, 0, cb_obs)

        cb_rel = NoHoverFocusComboBox()
        cb_rel.setItemDelegate(QStyledItemDelegate())
        cb_rel.addItems(CONS_RELATION_DICT.keys())
        cb_rel.setFixedWidth(64)
        cb_rel.setCurrentIndex(relation)
        self.setCellWidget(currentRow, 1, cb_rel)

        sb = QDoubleSpinBox()
        sb.setDecimals(decimals)
        sb.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        sb.installEventFilter(MouseWheelWidgetAdjustmentGuard(sb))
        default_value = threshold
        lb = default_value - 1e3
        ub = default_value + 1e3
        sb.setRange(lb, ub)
        sb.setStepType(QAbstractSpinBox.AdaptiveDecimalStepType)
        sb.setValue(default_value)
        self.setCellWidget(currentRow, 2, sb)

        btn_del = QPushButton(self.icon_trash, None)
        btn_del.setFixedSize(32, 24)
        btn_del.clicked.connect(self.remove_constraint_clicked)
        btn_del_container = QWidget()
        btn_del_layout = QHBoxLayout(btn_del_container)
        btn_del_layout.addWidget(btn_del)
        btn_del_layout.setAlignment(Qt.AlignLeft)
        btn_del_layout.setContentsMargins(2, 0, 0, 0)
        self.setCellWidget(self.rowCount() - 1, 4, btn_del_container)

        self.update_header_visibility()

    @pyqtSlot()
    def remove_constraint_clicked(self):
        self.update_header_visibility()
        button = self.sender()
        if button:
            row = self.indexAt(button.pos()).row()
            self.removeRow(row)

    def export_constraints(self) -> List[Tuple[str, int, float, bool, float]]:
        """
        Export the selected constraints.

        Returns
        -------
        List[Tuple[str, int, float, bool, float]]
            The list of constraints.
        """
        constraints_exported: List[Tuple[str, int, float, bool, float]] = []
        for i in range(self.rowCount()):
            constraints_exported.append(
                (
                    self.cellWidget(i, 0).currentText(),
                    CONS_RELATION_DICT[self.cellWidget(i, 1).currentText()],
                    self.cellWidget(i, 2).value(),
                    self.cellWidget(i, 3).isChecked(),
                    self.cellWidget(i, 2).decimals(),
                )
            )
        return constraints_exported

    def clear_constraints(self):
        self.setRowCount(0)
        self.update_header_visibility()



from importlib import resources
from typing import Any
from PyQt5.QtWidgets import QWidget, QAbstractSpinBox, QPushButton, QComboBox
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QLabel
from PyQt5.QtCore import Qt, QObject, QEvent, QSize
from PyQt5.QtGui import QIcon
import copy


def preventAnnoyingSpinboxScrollBehaviour(self, control: QAbstractSpinBox) -> None:
    control.setFocusPolicy(Qt.StrongFocus)
    control.installEventFilter(self.MouseWheelWidgetAdjustmentGuard(control))


class MouseWheelWidgetAdjustmentGuard(QObject):
    def __init__(self, parent: QObject):
        super().__init__(parent)

    def eventFilter(self, o: QObject, e: QEvent) -> bool:
        widget: QWidget = o
        if e.type() == QEvent.Wheel and not widget.hasFocus():
            e.ignore()
            return True
        return super().eventFilter(o, e)


def create_button(icon_file, tooltip, stylesheet=None, size=(32, 32), icon_size=None):
    icon_ref = resources.files(__package__) / f"./images/{icon_file}"
    with resources.as_file(icon_ref) as icon_path:
        icon = QIcon(str(icon_path))

    btn = QPushButton()
    if size:
        btn.setFixedSize(*size)
    btn.setIcon(icon)
    btn.setToolTip(tooltip)
    if icon_size:
        btn.setIconSize(QSize(*icon_size))

    if stylesheet is not None:
        btn.setStyleSheet(stylesheet)

    return btn


def filter_generator_config(name: str, config: dict[str, Any]):
    filtered_config: dict[str, Any] = {}
    if name == "neldermead":
        filtered_config["adaptive"] = config["adaptive"]
    elif name == "expected_improvement":
        filtered_config["turbo_controller"] = config["turbo_controller"]
        filtered_config["numerical_optimizer"] = config["numerical_optimizer"]
        filtered_config["max_travel_distances"] = config["max_travel_distances"]
    elif name == "upper_confidence_bound":
        filtered_config["turbo_controller"] = config["turbo_controller"]
        filtered_config["numerical_optimizer"] = config["numerical_optimizer"]
        filtered_config["max_travel_distances"] = config["max_travel_distances"]
        filtered_config["beta"] = config["beta"]
    elif name == "rcds":
        filtered_config["noise"] = config["noise"]
        filtered_config["step"] = config["step"]
    elif name == "mobo":
        filtered_config["numerical_optimizer"] = config["numerical_optimizer"]
        filtered_config["max_travel_distances"] = config["max_travel_distances"]
        filtered_config["reference_point"] = config["reference_point"]
    else:
        filtered_config = config

    return copy.deepcopy(filtered_config)


class NoHoverFocusComboBox(QComboBox):
    def focusInEvent(self, event):
        # Prevent focus if it's from a hover event
        if event.reason() == Qt.MouseFocusReason:
            event.ignore()
        else:
            super().focusInEvent(event)

    def mousePressEvent(self, event):
        # Ensure the combo box behaves normally when clicked
        self.setFocus()
        super().mousePressEvent(event)

    def wheelEvent(self, event):
        event.ignore()


class ModalOverlay(QDialog):
    def __init__(self, parent=None, info=None):
        super().__init__(parent)
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.Dialog)
        self.setGeometry(parent.geometry())

        layout = QVBoxLayout()
        if info is None:
            info = "Processing, please wait..."
        label = QLabel(info, self)
        label.setAlignment(Qt.AlignCenter)
        layout.addWidget(label)

        self.setLayout(layout)
        # Semi-transparent background
        self.setStyleSheet("background-color: rgba(0, 0, 0, 80);")



import pandas as pd
import pytest
from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtWidgets import QApplication


def test_routine_page_init(qtbot):
    from badger.gui.default.components.routine_page import BadgerRoutinePage

    window = BadgerRoutinePage()

    qtbot.addWidget(window)


def test_routine_generation(qtbot):
    from badger.errors import BadgerRoutineError
    from badger.utils import get_badger_version, get_xopt_version

    # test if a simple routine can be created
    from badger.gui.default.components.routine_page import BadgerRoutinePage

    window = BadgerRoutinePage()
    qtbot.addWidget(window)

    # Turn off relative to current
    window.env_box.relative_to_curr.setChecked(False)

    # test without anything selected
    with pytest.raises(BadgerRoutineError):
        window._compose_routine()

    # add generator -- still should raise error for no environment
    qtbot.keyClicks(window.generator_box.cb, "expected_improvement")
    with pytest.raises(BadgerRoutineError):
        window._compose_routine()

    # finally add the test environment
    qtbot.keyClicks(window.env_box.cb, "test")

    # click checkbox to select vars/objectives
    window.env_box.var_table.cellWidget(0, 0).setChecked(True)
    assert window.env_box.var_table.export_variables() == {"x0": [-1, 1]}

    window.env_box.obj_table.cellWidget(0, 0).setChecked(True)
    assert window.env_box.obj_table.export_objectives() == {"f": "MINIMIZE"}

    routine = window._compose_routine()
    assert routine.vocs.variables == {"x0": [-1, 1]}
    assert routine.vocs.objectives == {"f": "MINIMIZE"}
    assert routine.initial_points.empty

    # Test if badger and xopt version match with the current version
    assert routine.badger_version == get_badger_version()
    assert routine.xopt_version == get_xopt_version()


def test_add_additional_vars(qtbot):
    from badger.gui.default.components.routine_page import BadgerRoutinePage

    window = BadgerRoutinePage()
    qtbot.addWidget(window)

    # Turn off relative to current
    window.env_box.relative_to_curr.setChecked(False)

    # add generator
    qtbot.keyClicks(window.generator_box.cb, "expected_improvement")

    # add the test environment
    qtbot.keyClicks(window.env_box.cb, "test")

    # click checkbox to select vars/objectives
    window.env_box.var_table.cellWidget(0, 0).setChecked(True)
    assert window.env_box.var_table.export_variables() == {"x0": [-1, 1]}

    # Check that there is an extra row: X0 to X19, and one to enter a new PV
    n_rows = window.env_box.var_table.rowCount()
    assert n_rows == 21

    # Enter text in first cell of last row
    item = window.env_box.var_table.item(20, 1)
    item.setText("x20")
    assert window.env_box.var_table.item(20, 1).text() == "x20"

    # Send signal of table item changing
    window.env_box.var_table.cellChanged.emit(20, 1)

    # Why isn't this updating the table after changing the value?
    variables = {"x0": [-1, 1], "x20": [-1, 1]}

    # Check that new variable was added
    # Its checkbox checked by default when added
    assert window.env_box.var_table.addtl_vars == ["x20"]
    assert window.env_box.var_table.export_variables() == variables

    # Check that a new row was automatically added
    assert window.env_box.var_table.rowCount() == n_rows + 1


def test_initial_points(qtbot):
    # test to make sure initial points widget works properly
    from badger.gui.default.components.routine_page import BadgerRoutinePage

    window = BadgerRoutinePage()
    qtbot.addWidget(window)

    # Turn off relative to current
    window.env_box.relative_to_curr.setChecked(False)

    qtbot.keyClicks(window.env_box.cb, "test")
    qtbot.keyClicks(window.generator_box.cb, "random")

    window.env_box.var_table.cellWidget(0, 0).setChecked(True)
    window.env_box.var_table.cellWidget(1, 0).setChecked(True)
    window.env_box.var_table.cellWidget(2, 0).setChecked(True)

    assert window.env_box.init_table.horizontalHeader().count() == 3

    # test routine generation with fake current values selected
    qtbot.mouseClick(window.env_box.btn_add_curr, Qt.LeftButton)
    routine = window._compose_routine()
    assert (
        routine.initial_points.to_dict()
        == pd.DataFrame({"x0": 0.5, "x1": 0.5, "x2": 0.5}, index=[0]).to_dict()
    )


def test_ui_update(qtbot):
    # test to make sure initial points widget works properly
    from badger.gui.default.components.routine_page import BadgerRoutinePage
    from badger.tests.utils import create_routine

    window = BadgerRoutinePage()

    # test with none
    window.refresh_ui()

    routine = create_routine()
    window.refresh_ui(routine)
    idx = window.generators.index(routine.generator.name)
    window.select_generator(idx)

    assert window.generator_box.edit.toPlainText() == "{}\n"


def test_constraints(qtbot):
    # test if a simple routine can be created
    from badger.gui.default.components.routine_page import BadgerRoutinePage

    window = BadgerRoutinePage()
    qtbot.addWidget(window)

    qtbot.keyClicks(window.generator_box.cb, "expected_improvement")
    qtbot.keyClicks(window.env_box.cb, "test")

    # click checkbox to select vars/objectives
    window.env_box.var_table.cellWidget(0, 0).setChecked(True)
    window.env_box.obj_table.cellWidget(0, 0).setChecked(True)
    # Select constraint
    qtbot.mouseClick(window.env_box.btn_add_con, Qt.MouseButton.LeftButton)
    con_widget_name = window.env_box.con_table.cellWidget(0, 0)
    qtbot.keyClicks(con_widget_name, "c")
    con_widget_critical = window.env_box.con_table.cellWidget(0, 3)
    con_widget_critical.setChecked(True)

    routine = window._compose_routine()
    assert routine.vocs.constraints == {"c": ["GREATER_THAN", 0]}
    assert routine.critical_constraint_names == ["c"]


def test_observables(qtbot):
    # test if a simple routine can be created
    from badger.gui.default.components.routine_page import BadgerRoutinePage

    window = BadgerRoutinePage()
    qtbot.addWidget(window)

    qtbot.keyClicks(window.generator_box.cb, "expected_improvement")
    qtbot.keyClicks(window.env_box.cb, "test")

    # click checkbox to select vars/objectives
    window.env_box.var_table.cellWidget(0, 0).setChecked(True)
    window.env_box.obj_table.cellWidget(0, 0).setChecked(True)
    # Select observable
    qtbot.mouseClick(window.env_box.btn_add_sta, Qt.MouseButton.LeftButton)
    obs_item = window.env_box.list_obs.item(0)
    obs_widget = window.env_box.list_obs.itemWidget(obs_item)
    qtbot.keyClicks(obs_widget.cb_sta, "c")

    routine = window._compose_routine()
    assert routine.vocs.observables == ["c"]


def test_add_random_points(qtbot):
    # test to add random points to initial points table
    from badger.gui.default.components.routine_page import BadgerRoutinePage

    window = BadgerRoutinePage()
    qtbot.addWidget(window)

    # Turn off relative to current
    window.env_box.relative_to_curr.setChecked(False)

    qtbot.keyClicks(window.env_box.cb, "test")
    qtbot.keyClicks(window.generator_box.cb, "random")

    window.env_box.var_table.cellWidget(0, 0).setChecked(True)
    window.env_box.var_table.cellWidget(1, 0).setChecked(True)
    window.env_box.var_table.cellWidget(2, 0).setChecked(True)

    def handle_dialog():
        while window.rc_dialog is None:
            QApplication.processEvents()

        # Set number of points to 5, frac to 0.05, then add them
        window.rc_dialog.sb_np.setValue(5)
        window.rc_dialog.sb_frac.setValue(0.05)
        qtbot.mouseClick(window.rc_dialog.btn_add, Qt.MouseButton.LeftButton)

    QTimer.singleShot(0, handle_dialog)

    qtbot.mouseClick(window.env_box.btn_add_rand, Qt.LeftButton)
    routine = window._compose_routine()

    assert routine.initial_points.shape[0] == 5
    # curr is 0.5, frac is 0.05, range is [-1, 1]
    # so max value should be 0.6, min value should be 0.4
    assert routine.initial_points.to_numpy().max() <= 0.6
    assert routine.initial_points.to_numpy().min() >= 0.4


# TODO: Test if the EI, Simplex, and RCDS params show o the params editor
# are the simplified versions
def test_simplified_generator_params(qtbot):
    pass


# TODO: First load an old routine w/ initail points,
# then create a new routine and check if the initial points panel
# is cleared
def test_initial_points_clear_when_create_routine(qtbot):
    pass


# TODO: Test if env selector reacts to scroll events, it should not
def test_scroll_on_environment_selector(qtbot):
    pass


# TODO: Test if generator selector reacts to scroll events, it should not
def test_scroll_on_generator_selector(qtbot):
    pass


# TODO: Test relative to current behavior, including the auto calculated
# bounds and initial points wrt the current variable values
def test_relative_to_current(qtbot):
    pass



from PyQt5.QtWidgets import QWidget, QHBoxLayout
from PyQt5.QtWidgets import QToolButton, QMenu, QAction
from PyQt5.QtGui import QIcon, QFont
from PyQt5.QtCore import pyqtSignal
from importlib import resources
from badger.gui.default.utils import create_button

stylesheet_del = """
QPushButton:hover:pressed
{
    background-color: #C7737B;
}
QPushButton:hover
{
    background-color: #BF616A;
}
QPushButton
{
    background-color: #A9444E;
}
"""

stylesheet_log = """
QPushButton:hover:pressed
{
    background-color: #88C0D0;
}
QPushButton:hover
{
    background-color: #72A4B4;
}
QPushButton
{
    background-color: #5C8899;
    color: #000000;
}
"""

stylesheet_ext = """
QPushButton:hover:pressed
{
    background-color: #4DB6AC;
}
QPushButton:hover
{
    background-color: #26A69A;
}
QPushButton
{
    background-color: #00897B;
}
"""

stylesheet_run = """
QToolButton:hover:pressed
{
    background-color: #92D38C;
}
QToolButton:hover
{
    background-color: #6EC566;
}
QToolButton
{
    background-color: #4AB640;
    color: #000000;
}
"""

stylesheet_stop = """
QToolButton:hover:pressed
{
    background-color: #C7737B;
}
QToolButton:hover
{
    background-color: #BF616A;
}
QToolButton
{
    background-color: #A9444E;
}
"""


class BadgerActionBar(QWidget):
    sig_start = pyqtSignal()
    sig_start_until = pyqtSignal()
    sig_stop = pyqtSignal()

    sig_delete_run = pyqtSignal()
    sig_logbook = pyqtSignal()
    sig_reset_env = pyqtSignal()
    sig_jump_to_optimal = pyqtSignal()
    sig_dial_in = pyqtSignal()
    sig_ctrl = pyqtSignal(bool)
    sig_open_extensions_palette = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.init_ui()
        self.config_logic()

    def init_ui(self):
        icon_ref = resources.files(__package__) / "../images/play.png"
        with resources.as_file(icon_ref) as icon_path:
            self.icon_play = QIcon(str(icon_path))
        icon_ref = resources.files(__package__) / "../images/pause.png"
        with resources.as_file(icon_ref) as icon_path:
            self.icon_pause = QIcon(str(icon_path))
        icon_ref = resources.files(__package__) / "../images/stop.png"
        with resources.as_file(icon_ref) as icon_path:
            self.icon_stop = QIcon(str(icon_path))

        hbox_action = QHBoxLayout(self)
        hbox_action.setContentsMargins(0, 0, 0, 0)

        # Background widget
        self.bg = QWidget()
        hbox_action.addWidget(self.bg, 1)
        self.bg.setObjectName("ActionBar")
        hbox_bg = QHBoxLayout(self.bg)
        hbox_bg.setContentsMargins(8, 8, 8, 8)

        cool_font = QFont()
        cool_font.setWeight(QFont.DemiBold)
        cool_font.setPixelSize(13)

        self.btn_del = create_button("trash.png", "Delete run", stylesheet_del)
        self.btn_log = create_button("book.png", "Logbook", stylesheet_log)
        self.btn_reset = create_button("undo.png", "Reset environment")
        self.btn_opt = create_button("star.png", "Jump to optimum")
        self.btn_set = create_button("set.png", "Dial in solution")
        self.btn_ctrl = create_button("pause.png", "Pause")
        self.btn_ctrl._status = "pause"

        self.btn_del.setDisabled(True)
        self.btn_log.setDisabled(True)
        self.btn_reset.setDisabled(True)
        self.btn_opt.setDisabled(True)
        self.btn_set.setDisabled(True)
        self.btn_ctrl.setDisabled(True)

        # self.btn_stop = btn_stop = QPushButton('Run')
        self.btn_stop = QToolButton()
        self.btn_stop.setFixedSize(96, 32)
        self.btn_stop.setFont(cool_font)
        self.btn_stop.setStyleSheet(stylesheet_run)

        # add button for extensions
        self.btn_open_extensions_palette = btn_extensions = create_button(
            "extension.png", "Open extensions", stylesheet_ext
        )

        # Create a menu and add options
        self.run_menu = menu = QMenu(self)
        menu.setFixedWidth(128)
        self.run_action = run_action = QAction("Run", self)
        run_action.setIcon(self.icon_play)
        self.run_until_action = run_until_action = QAction("Run until", self)
        run_until_action.setIcon(self.icon_play)
        menu.addAction(run_action)
        menu.addAction(run_until_action)

        # Set the menu as the run button's dropdown menu
        self.btn_stop.setMenu(menu)
        self.btn_stop.setDefaultAction(run_action)
        self.btn_stop.setPopupMode(QToolButton.MenuButtonPopup)
        self.btn_stop.setDisabled(False)
        # btn_stop.setToolTip('')

        # Config button
        self.btn_config = btn_config = create_button("tools.png", "Configure run")
        btn_config.hide()

        hbox_bg.addWidget(self.btn_del)
        # hbox_action.addWidget(btn_edit)
        hbox_bg.addWidget(self.btn_log)
        hbox_bg.addStretch(1)
        hbox_bg.addWidget(self.btn_reset)
        hbox_bg.addWidget(self.btn_ctrl)
        hbox_bg.addWidget(self.btn_stop)
        hbox_bg.addWidget(self.btn_opt)
        hbox_bg.addWidget(self.btn_set)
        hbox_bg.addStretch(1)
        hbox_bg.addWidget(btn_extensions)
        hbox_bg.addWidget(btn_config)

        self.setStyleSheet("""
            #ActionBar {
                background-color: #455364;
            }
        """)

    def config_logic(self):
        self.btn_del.clicked.connect(self.delete_run)
        self.btn_log.clicked.connect(self.logbook)
        self.btn_reset.clicked.connect(self.reset_env)
        self.btn_opt.clicked.connect(self.jump_to_optimal)
        self.btn_set.clicked.connect(self.dial_in)
        self.btn_ctrl.clicked.connect(self.ctrl_routine)
        self.run_action.triggered.connect(self.set_run_action)
        self.run_until_action.triggered.connect(self.set_run_until_action)
        self.btn_open_extensions_palette.clicked.connect(self.open_extensions_palette)

    def lock(self):
        self.btn_del.setDisabled(True)
        self.btn_log.setDisabled(True)
        self.btn_reset.setDisabled(True)
        self.btn_ctrl.setDisabled(True)
        self.btn_stop.setDisabled(True)
        self.btn_opt.setDisabled(True)
        self.btn_set.setDisabled(True)

    def unlock(self):
        self.btn_del.setDisabled(False)
        self.btn_log.setDisabled(False)
        self.btn_reset.setDisabled(False)
        self.btn_ctrl.setDisabled(False)
        self.btn_stop.setDisabled(False)
        self.btn_opt.setDisabled(False)
        self.btn_set.setDisabled(False)

    def routine_invalid(self):
        self.btn_stop.setDisabled(False)

    def routine_finished(self):
        self.btn_ctrl.setIcon(self.icon_pause)
        self.btn_ctrl.setToolTip("Pause")
        self.btn_ctrl._status = "pause"
        self.btn_ctrl.setDisabled(True)

        # Note the order of the following two lines cannot be changed!
        self.btn_stop.setPopupMode(QToolButton.MenuButtonPopup)
        self.btn_stop.setStyleSheet(stylesheet_run)
        self.run_action.setText("Run")
        self.run_action.setIcon(self.icon_play)
        self.run_until_action.setText("Run until")
        self.run_until_action.setIcon(self.icon_play)
        # self.btn_stop.setToolTip('')
        self.btn_stop.setDisabled(False)

        self.btn_reset.setDisabled(False)
        self.btn_set.setDisabled(False)
        self.btn_del.setDisabled(False)

    def toggle_reset(self, locked):
        self.btn_reset.setDisabled(locked)

    def toggle_run(self, locked):
        self.btn_stop.setDisabled(locked)

    def toggle_other(self, locked):
        self.btn_del.setDisabled(locked)
        self.btn_log.setDisabled(locked)
        self.btn_opt.setDisabled(locked)
        self.btn_set.setDisabled(locked)

    def run_start(self):
        self.btn_stop.setStyleSheet(stylesheet_stop)
        self.btn_stop.setPopupMode(QToolButton.DelayedPopup)
        self.btn_stop.setDisabled(False)
        self.run_action.setText("Stop")
        self.run_action.setIcon(self.icon_stop)
        self.run_until_action.setText("Stop")
        self.run_until_action.setIcon(self.icon_stop)
        self.btn_ctrl.setDisabled(False)
        self.btn_set.setDisabled(True)

    def set_run_action(self):
        if self.btn_stop.defaultAction() is not self.run_action:
            self.btn_stop.setDefaultAction(self.run_action)

        if self.run_action.text() == "Run":
            self.btn_stop.setDisabled(True)
            self.sig_start.emit()
        else:
            self.btn_stop.setDisabled(True)
            self.sig_stop.emit()

    def set_run_until_action(self):
        if self.btn_stop.defaultAction() is not self.run_until_action:
            self.btn_stop.setDefaultAction(self.run_until_action)

        if self.run_until_action.text() == "Run until":
            self.sig_start_until.emit()
        else:
            self.btn_stop.setDisabled(True)
            self.sig_stop.emit()

    def delete_run(self):
        self.sig_delete_run.emit()

    def logbook(self):
        self.sig_logbook.emit()

    def reset_env(self):
        self.sig_reset_env.emit()

    def jump_to_optimal(self):
        self.sig_jump_to_optimal.emit()

    def dial_in(self):
        self.sig_dial_in.emit()

    def ctrl_routine(self):
        if self.btn_ctrl._status == "pause":
            self.sig_ctrl.emit(True)
            self.btn_ctrl.setIcon(self.icon_play)
            self.btn_ctrl.setToolTip("Resume")
            self.btn_ctrl._status = "play"
        else:
            self.sig_ctrl.emit(False)
            self.btn_ctrl.setIcon(self.icon_pause)
            self.btn_ctrl.setToolTip("Pause")
            self.btn_ctrl._status = "pause"

    def open_extensions_palette(self):
        self.sig_open_extensions_palette.emit()

    def env_ready(self):
        self.btn_log.setDisabled(False)
        self.btn_opt.setDisabled(False)



import traceback

from PyQt5.QtWidgets import (
    QMainWindow,
    QPushButton,
    QVBoxLayout,
    QWidget,
    QLabel,
    QMessageBox,
    QSizePolicy,
)
from PyQt5.QtCore import Qt
from badger.gui.default.components.analysis_extensions import (
    AnalysisExtension,
    ParetoFrontViewer,
    BOVisualizer,
)


class ExtensionsPalette(QMainWindow):
    """
    A QMainWindow-based user interface for managing and launching extensions in Badger.

    Parameters
    ----------
    run_monitor : RunMonitor
        The run monitor associated with the palette.

    Attributes
    ----------
    base_text : str
        Base text for the number of active extensions.
    text_box : QLabel
        QLabel widget for displaying the number of active extensions.
    btn_data_viewer : QPushButton
        QPushButton for launching the ParetoFrontViewer extension.

    Methods
    -------
    n_active_extensions
        Property to get the number of active extensions.
    update_palette
        Update the display of the active extensions count.
    add_pf_viewer
        Open the ParetoFrontViewer extension.
    add_child_window_to_monitor(child_window)
        Add a child window to the run monitor.

    """

    def __init__(self, run_monitor):
        """
        Initialize the ExtensionsPalette.

        Parameters
        ----------
        run_monitor : RunMonitor
            The run monitor associated with the palette.

        """
        super().__init__(parent=run_monitor)

        self.run_monitor = run_monitor

        self.setWindowTitle("Badger Extensions Palette")
        self.resize(320, 240)

        central_widget = QWidget()
        self.setCentralWidget(central_widget)

        layout = QVBoxLayout()

        self.base_text = "Number of active exensions: "
        self.text_box = QLabel(self.base_text + "0", self)
        self.text_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.text_box.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.btn_data_viewer = QPushButton("ParetoFrontViewer")
        self.btn_bo_visualizer = QPushButton("BOVisualizer")

        layout.addWidget(self.btn_data_viewer)
        layout.addWidget(self.btn_bo_visualizer)
        layout.addStretch()
        layout.addWidget(self.text_box)

        central_widget.setLayout(layout)

        self.btn_data_viewer.clicked.connect(self.add_pf_viewer)
        self.btn_bo_visualizer.clicked.connect(self.add_bo_visualizer)

    @property
    def n_active_extensions(self):
        """
        Property to get the number of active extensions.

        Returns
        -------
        int
            The number of active extensions.

        """
        return len(self.run_monitor.active_extensions)

    def update_palette(self):
        self.text_box.setText(self.base_text + str(self.n_active_extensions))

    def add_pf_viewer(self):
        """
        Open the ParetoFrontViewer extension.

        """
        self.add_child_window_to_monitor(ParetoFrontViewer())

    def add_bo_visualizer(self):
        """
        Open the BOVisualizer extension.

        """
        self.add_child_window_to_monitor(BOVisualizer())

    def add_child_window_to_monitor(self, child_window: AnalysisExtension):
        """
        Add a child window to the run monitor.

        Parameters
        ----------
        child_window : AnalysisExtension
            The child window (extension) to add to the run monitor.

        """
        child_window.window_closed.connect(self.run_monitor.extension_window_closed)
        self.run_monitor.active_extensions.append(child_window)

        try:
            if self.run_monitor.routine is not None:
                child_window.update_window(self.run_monitor.routine)
        except ValueError:
            QMessageBox.warning(
                self, "Extension is not applicable!", traceback.format_exc()
            )
            self.run_monitor.active_extensions.remove(child_window)
            return

        child_window.show()

        self.update_palette()



from PyQt5.QtWidgets import (
    QHBoxLayout,
    QPushButton,
    QWidget,
    QDoubleSpinBox,
    QAbstractSpinBox,
)
from PyQt5.QtWidgets import QComboBox, QCheckBox, QStyledItemDelegate
from PyQt5.QtCore import Qt
from badger.gui.default.utils import (
    MouseWheelWidgetAdjustmentGuard,
    NoHoverFocusComboBox,
)


def constraint_item(
    options, remove_item, name=None, relation=0, threshold=0, critical=False, decimals=4
):
    # relation: 0 for >, 1 for <, 2 for =
    widget = QWidget()
    hbox = QHBoxLayout(widget)
    hbox.setContentsMargins(2, 2, 2, 2)
    # hbox.setSpacing(0)
    widget.cb_obs = cb_obs = NoHoverFocusComboBox()
    cb_obs.setFixedWidth(200)
    cb_obs.setItemDelegate(QStyledItemDelegate())
    cb_obs.addItems(options)
    try:
        idx = options.index(name)
    except:
        idx = 0
    cb_obs.setCurrentIndex(idx)

    widget.cb_rel = cb_rel = QComboBox()
    cb_rel.setItemDelegate(QStyledItemDelegate())
    cb_rel.addItems([">", "<", "="])
    cb_rel.setFixedWidth(64)
    cb_rel.setCurrentIndex(relation)

    widget.sb = sb = QDoubleSpinBox()
    sb.setDecimals(decimals)
    sb.setFocusPolicy(Qt.StrongFocus)
    sb.installEventFilter(MouseWheelWidgetAdjustmentGuard(sb))
    default_value = threshold
    lb = default_value - 1e3
    ub = default_value + 1e3
    sb.setRange(lb, ub)
    sb.setStepType(QAbstractSpinBox.AdaptiveDecimalStepType)
    sb.setValue(default_value)

    widget.check_crit = check_crit = QCheckBox("Critical")
    check_crit.setChecked(critical)

    widget.btn_del = btn_del = QPushButton("Remove")
    btn_del.setFixedSize(72, 24)
    btn_del.hide()

    hbox.addWidget(check_crit)
    hbox.addWidget(cb_obs)
    hbox.addWidget(cb_rel)
    hbox.addWidget(sb, 1)
    hbox.addWidget(btn_del)

    btn_del.clicked.connect(remove_item)

    def show_button(event):
        btn_del.show()

    def hide_button(event):
        btn_del.hide()

    widget.enterEvent = show_button
    widget.leaveEvent = hide_button

    return widget



import os

import pytest


class TestRoutine:
    def test_routine_init(self):
        from badger.tests.utils import create_routine

        routine = create_routine()
        routine.evaluate_data(routine.initial_points)

        assert len(routine.data) == 1

    def test_routine_serialization(self):
        from badger.routine import Routine
        from badger.tests.utils import create_routine

        routine = create_routine()
        routine.evaluate_data(routine.initial_points)
        routine.dump("test.yaml")

        lroutine = Routine.from_file("test.yaml")
        assert lroutine.environment.variable_names == routine.environment.variable_names
        lroutine.evaluate_data(routine.initial_points)

        print(lroutine.data)

        assert len(lroutine.data) == 2

    def test_routine_env_dump(self):
        from badger.environment import Environment
        from badger.routine import Routine
        from badger.tests.utils import create_routine

        routine = create_routine()
        assert isinstance(routine.environment, Environment)

        routine.environment.flag = 1
        assert routine.environment.flag == 1

        routine_str = routine.yaml()
        routine_re = Routine.from_yaml(routine_str)
        assert routine_re.environment.flag == 1

    @pytest.fixture(scope="module", autouse=True)
    def clean_up(self):
        yield
        files = ["test.yaml"]
        for f in files:
            if os.path.exists(f):
                os.remove(f)



import multiprocessing

import pytest
from PyQt5.QtCore import QEventLoop, Qt, QTimer
from PyQt5.QtTest import QSignalSpy
from PyQt5.QtWidgets import QApplication
from unittest.mock import Mock


class TestRoutineRunner:
    @pytest.fixture(scope="session")
    def init_multiprocessing(self):
        multiprocessing.set_start_method("fork", force=True)

    @pytest.fixture(scope="session")
    def init_multiprocessing_alt(self):
        # Use 'spawn' to start new processes instead of 'fork'
        multiprocessing.set_start_method("spawn", force=True)

    @pytest.fixture
    def process_manager(self):
        from badger.gui.default.components.create_process import CreateProcess
        from badger.gui.default.components.process_manager import ProcessManager

        process_manager = ProcessManager()
        process_builder = CreateProcess()
        process_builder.subprocess_prepared.connect(process_manager.add_to_queue)
        process_builder.create_subprocess()

        yield process_manager

        process_manager.close_proccesses()

    @pytest.fixture
    def instance(self, process_manager, init_multiprocessing):
        from badger.archive import save_tmp_run
        from badger.gui.default.components.routine_runner import (
            BadgerRoutineSubprocess,
        )
        from badger.tests.utils import create_routine

        routine = create_routine()
        _ = save_tmp_run(routine)
        instance = BadgerRoutineSubprocess(process_manager, routine)
        instance.pause_event = multiprocessing.Event()
        return instance

    def test_ctrl_routine(self, instance) -> None:
        # Test setting the event
        instance.ctrl_routine(True)
        assert not instance.pause_event.is_set()

        # Test clearing the event
        instance.ctrl_routine(False)
        assert instance.pause_event.is_set()

    def test_stop_routine(self, instance):
        instance.run()
        instance.stop_routine()
        assert instance.stop_event.is_set()

    def test_save_init_vars(self, instance):
        sig_env_ready_spy = QSignalSpy(instance.signals.env_ready)
        instance.save_init_vars()
        assert len(sig_env_ready_spy) == 1

    def test_after_evaluate(self, instance):
        sig_progress_spy = QSignalSpy(instance.signals.progress)
        instance.setup_timer()
        instance.after_evaluate(True)
        assert len(sig_progress_spy) == 1
        instance.timer.stop()

    def test_check_queue(self, instance):
        sig_finished_spy = QSignalSpy(instance.signals.finished)
        instance.run()
        instance.data_and_error_queue.empty = Mock(return_value=True)
        instance.check_queue()
        instance.stop_routine()
        assert len(sig_finished_spy) == 1
        assert not instance.timer.isActive()

        # sig_progress_spy = QSignalSpy(instance.signals.progress)
        # instance.run()
        # nstance.check_queue()
        # instance.stop_routine()
        # assert len(sig_progress_spy) > 0

    def test_setup_timer(self, instance):
        instance.setup_timer()
        assert isinstance(instance.timer, QTimer)
        assert instance.timer.interval() == 100
        assert instance.timer.isActive()

        instance.timer.stop()

    def test_run(self, instance):
        instance.run()
        instance.ctrl_routine(True)
        assert instance.routine_process.is_alive()
        assert instance.wait_event.is_set()
        instance.stop_routine()

    def test_set_termination_condition(self, instance):
        instance.set_termination_condition(True)
        assert instance.termination_condition

    # TODO: check for signal emit message

    def test_turbo_with_routine_runner(self, qtbot, init_multiprocessing_alt):
        # TODO: make this test more stable
        return

        from badger.gui.acr.windows.main_window import BadgerMainWindow
        from badger.gui.default.windows.message_dialog import (
            BadgerScrollableMessageBox,
        )

        window = BadgerMainWindow()
        # qtbot.addWidget(window)

        loop = QEventLoop()
        QTimer.singleShot(1000, loop.quit)  # 1000 ms pause
        loop.exec_()

        # Create and save a routine
        editor = window.home_page.routine_editor
        # Turn off relative to current
        editor.routine_page.env_box.relative_to_curr.setChecked(False)
        # Config env and vocs
        qtbot.keyClicks(editor.routine_page.env_box.cb, "test")
        editor.routine_page.env_box.var_table.cellWidget(0, 0).setChecked(True)
        editor.routine_page.env_box.obj_table.cellWidget(0, 0).setChecked(True)
        qtbot.mouseClick(
            editor.routine_page.env_box.btn_add_curr, Qt.MouseButton.LeftButton
        )

        # Run the routine
        monitor = window.home_page.run_monitor
        monitor.testing = True

        def handle_dialog():
            while monitor.tc_dialog is None:
                QApplication.processEvents()

            # Set max evaluation to 2, then run the optimization
            monitor.tc_dialog.sb_max_eval.setValue(2)
            qtbot.mouseClick(monitor.tc_dialog.btn_run, Qt.MouseButton.LeftButton)

        QTimer.singleShot(0, handle_dialog)

        # Bypass the possible Badger error dialog
        def patched_showEvent(original_showEvent):
            def inner(ins, event):
                original_showEvent(ins, event)  # Call the original showEvent
                QTimer.singleShot(100, ins.accept)  # Close the dialog after 100 ms

            return inner

        BadgerScrollableMessageBox.showEvent = patched_showEvent(
            BadgerScrollableMessageBox.showEvent
        )

        action_bar = window.home_page.run_action_bar
        action_bar.run_until_action.trigger()
        monitor.routine_runner.data_and_error_queue.empty = Mock(return_value=True)

        # Wait until the run is done
        while monitor.running:
            qtbot.wait(100)

        assert len(monitor.routine.data) == 2

        window.process_manager.close_proccesses()



from PyQt5.QtWidgets import QHBoxLayout, QPushButton, QWidget
from PyQt5.QtWidgets import QStyledItemDelegate
from badger.gui.default.utils import NoHoverFocusComboBox


def state_item(options, remove_item, name=None):
    widget = QWidget()
    hbox = QHBoxLayout(widget)
    hbox.setContentsMargins(2, 2, 2, 2)
    # hbox.setSpacing(0)
    widget.cb_sta = cb_sta = NoHoverFocusComboBox()
    # cb_sta.setFixedWidth(200)
    cb_sta.setItemDelegate(QStyledItemDelegate())
    cb_sta.addItems(options)
    try:
        idx = options.index(name)
    except:
        idx = 0
    cb_sta.setCurrentIndex(idx)

    widget.btn_del = btn_del = QPushButton("Remove")
    btn_del.setFixedSize(72, 24)
    btn_del.hide()

    hbox.addWidget(cb_sta, 1)
    hbox.addWidget(btn_del)

    btn_del.clicked.connect(remove_item)

    def show_button(event):
        btn_del.show()

    def hide_button(event):
        btn_del.hide()

    widget.enterEvent = show_button
    widget.leaveEvent = hide_button

    return widget



from functools import wraps
from typing import Callable, Optional, ParamSpec, cast
from PyQt5.QtWidgets import QHBoxLayout, QWidget, QVBoxLayout
from PyQt5.QtWidgets import QSizePolicy

from badger.gui.default.components.bo_visualizer.types import ConfigurableOptions
from badger.routine import Routine
from .ui_components import UIComponents
from .plotting_area import PlottingArea
from .model_logic import ModelLogic
from PyQt5.QtCore import Qt
from xopt.generators.bayesian.bayesian_generator import BayesianGenerator

import logging

logger = logging.getLogger(__name__)

Param = ParamSpec("Param")


def signal_logger(text: str):
    def decorator(fn: Callable[Param, None]) -> Callable[Param, None]:
        @wraps(fn)
        def wrapper(*args: Param.args, **kwargs: Param.kwargs):
            logger.debug(f"{text}")
            return fn(*args, **kwargs)

        return wrapper

    return decorator


DEFAULT_PARAMETERS: ConfigurableOptions = {
    "plot_options": {
        "n_grid": 50,
        "n_grid_range": (10, 100),
        "show_samples": True,
        "show_prior_mean": False,
        "show_feasibility": False,
        "show_acq_func": True,
    },
    "variable_1": 0,
    "variable_2": 1,
    "include_variable_2": True,
}


class BOPlotWidget(QWidget):
    config_parameters: ConfigurableOptions

    def __init__(
        self, parent: Optional[QWidget] = None, routine: Optional[Routine] = None
    ):
        logger.debug("Initializing BOPlotWidget")
        super().__init__(parent)

        # Set default configuration parameters
        self.config_parameters = DEFAULT_PARAMETERS

        self.selected_variables: list[str] = []  # Initialize selected_variables

        # Initialize model logic and UI components with None or default values
        self.model_logic = ModelLogic(routine, routine.vocs if routine else None)
        self.ui_components = UIComponents(
            self.config_parameters, routine.vocs if routine else None
        )
        self.plotting_area = PlottingArea()

        main_layout = QHBoxLayout(self)
        controls_layout = QVBoxLayout()

        controls_layout.addLayout(self.ui_components.create_axis_layout())
        controls_layout.addWidget(self.ui_components.create_reference_inputs())
        controls_layout.addWidget(self.ui_components.create_options_section())
        controls_layout.addLayout(self.ui_components.create_buttons())

        main_layout.addLayout(controls_layout)
        main_layout.addWidget(self.plotting_area, stretch=1)

        self.setLayout(main_layout)

        ExpandingPolicy = QSizePolicy.Policy.Expanding

        self.setSizePolicy(ExpandingPolicy, ExpandingPolicy)
        self.setMinimumSize(1250, 720)

    def initialize_widget(
        self, routine: Routine, update_routine: Callable[[Routine], None]
    ) -> None:
        logger.debug("Initializing plot with routine")
        self.model_logic.update_routine(routine)
        logger.debug("Update vocs in UI components")
        self.ui_components.update_vocs(routine.vocs)
        self.ui_components.update_variables(self.config_parameters)

        # Set up connections
        logger.debug("Setting up connections")
        self.setup_connections(routine, update_routine)

        # Initialize UI Components
        self.ui_components.initialize_ui_components(self.config_parameters)

        # Trigger the axis selection changed to disable reference points for default selected variables
        logger.debug("Triggering axis selection changed")
        self.on_axis_selection_changed()

    def setup_connections(
        self, routine: Routine, update_routine: Callable[[Routine], None]
    ) -> None:
        # Disconnect existing connections
        try:
            self.ui_components.update_button.clicked.disconnect()
        except TypeError:
            pass  # No connection to disconnect

        self.ui_components.update_button.clicked.connect(
            lambda: signal_logger("Update button clicked")(
                lambda: update_routine(routine)
            )()
        )

        # Similarly for other signals
        try:
            self.ui_components.x_axis_combo.currentIndexChanged.disconnect()
        except TypeError:
            pass
        self.ui_components.x_axis_combo.currentIndexChanged.connect(
            lambda: signal_logger("Updated 'x_axis_combo'")(
                lambda: self.on_axis_selection_changed()
            )()
        )

        try:
            self.ui_components.y_axis_combo.currentIndexChanged.disconnect()
        except TypeError:
            pass
        self.ui_components.y_axis_combo.currentIndexChanged.connect(
            lambda: signal_logger("Updated 'y_axis_combo'")(
                lambda: self.on_axis_selection_changed()
            )()
        )

        try:
            self.ui_components.y_axis_checkbox.stateChanged.disconnect()
        except TypeError:
            pass
        self.ui_components.y_axis_checkbox.stateChanged.connect(
            lambda: signal_logger("Updated 'y_axis_checkbox'")(
                lambda: self.on_axis_selection_changed()
            )()
        )

        # Plot options
        plot_options_checkboxes = [
            self.ui_components.acq_func_checkbox,
            self.ui_components.show_samples_checkbox,
            self.ui_components.show_prior_mean_checkbox,
            self.ui_components.show_feasibility_checkbox,
        ]

        for checkbox in plot_options_checkboxes:
            logger.debug(f"Setting up connection for checkbox: {checkbox.text()}")
            try:
                checkbox.stateChanged.disconnect()
            except TypeError:
                pass
            checkbox.stateChanged.connect(
                lambda: signal_logger("Updated checkbox")(lambda: self.update_plot())()
            )

        # No. of Grid Points
        try:
            self.ui_components.n_grid.valueChanged.disconnect()
        except TypeError:
            pass
        self.ui_components.n_grid.valueChanged.connect(
            lambda: signal_logger("Updated 'n_grid' spinbox")(
                lambda: self.update_plot()
            )()
        )

        # Reference inputs

        if self.ui_components.reference_table is not None:
            try:
                self.ui_components.reference_table.cellChanged.disconnect()
            except TypeError:
                pass
            self.ui_components.reference_table.cellChanged.connect(
                lambda: signal_logger("Updated 'reference_table'")(
                    lambda: self.update_plot()
                )()
            )

    def on_axis_selection_changed(self):
        if not self.model_logic.vocs or not self.ui_components.ref_inputs:
            # vocs or ref_inputs is not yet set; skip processing
            return

        logger.debug("Axis selection changed")

        selected_variables: list[str] = []

        previous_selected_options = (
            self.config_parameters["variable_1"],
            self.config_parameters["variable_2"],
            self.config_parameters["include_variable_2"],
        )

        x_var_index = self.ui_components.x_axis_combo.currentIndex()
        self.config_parameters["variable_1"] = x_var_index
        x_var_text = self.ui_components.x_axis_combo.currentText()
        if x_var_text != "":
            selected_variables.append(x_var_text)

        include_var2 = self.ui_components.y_axis_checkbox.isChecked()
        self.config_parameters["include_variable_2"] = include_var2
        if include_var2:
            y_var_index = self.ui_components.y_axis_combo.currentIndex()
            self.config_parameters["variable_2"] = y_var_index
            y_var_text = self.ui_components.y_axis_combo.currentText()
            if y_var_text != "":
                selected_variables.append(y_var_text)

        current_selected_options = (
            self.config_parameters["variable_1"],
            self.config_parameters["variable_2"],
            self.config_parameters["include_variable_2"],
        )

        logger.debug(f"previous_selected_options: {previous_selected_options}")
        logger.debug(f"current_selected_options: {current_selected_options}")

        # Update the selected variables based on the selected indices
        self.selected_variables = list(set(selected_variables))

        self.ui_components.x_axis_combo.setCurrentIndex(
            self.config_parameters["variable_1"]
        )
        self.ui_components.y_axis_combo.setCurrentIndex(
            self.config_parameters["variable_2"]
        )

        if previous_selected_options != current_selected_options:
            print("Selected variables for plotting:", self.selected_variables)
            # Update the reference point table based on the selected variables
            if self.ui_components.reference_table is not None:
                self.ui_components.reference_table.blockSignals(True)
                self.update_reference_point_table(self.selected_variables)
                self.ui_components.reference_table.blockSignals(False)
            # Only update plot if the selection has changed
            self.update_plot()

    def update_plot(
        self, interval: Optional[float] = None, requires_rebuild: bool = False
    ) -> None:
        logger.debug("Updating plot in BOPlotWidget")
        if not self.model_logic.routine or not self.model_logic.vocs:
            print("Cannot update plot: routine or vocs are not available.")
            return

        # Ensure selected_variables is not empty
        if len(self.selected_variables) == 0:
            logger.error("No variables selected for plotting")
            return

        # **Add validation for the number of variables**
        if len(self.selected_variables) > 2:
            logger.error("Too many variables selected for plotting")
            return

        for var in self.selected_variables:
            if var == "":
                logger.error("Empty variable selected for plotting")
                return

        n_grid_value = self.ui_components.n_grid.value()

        if n_grid_value < self.config_parameters["plot_options"]["n_grid_range"][0]:
            logger.error(
                f"Number of grid points is less than the minimum value: {n_grid_value}"
            )
            return

        # Update the plot options
        self.config_parameters["plot_options"]["n_grid"] = (
            self.ui_components.n_grid.value()
        )
        self.config_parameters["plot_options"]["show_samples"] = (
            self.ui_components.show_samples_checkbox.isChecked()
        )
        self.config_parameters["plot_options"]["show_prior_mean"] = (
            self.ui_components.show_prior_mean_checkbox.isChecked()
        )
        self.config_parameters["plot_options"]["show_feasibility"] = (
            self.ui_components.show_feasibility_checkbox.isChecked()
        )
        self.config_parameters["plot_options"]["show_acq_func"] = (
            self.ui_components.acq_func_checkbox.isChecked()
        )

        # Disable signals for the reference table to prevent updating the plot multiple times
        if self.ui_components.reference_table is not None:
            self.ui_components.reference_table.blockSignals(True)

            # Disable and gray out the reference points for selected variables
            self.update_reference_point_table(self.selected_variables)

            self.ui_components.reference_table.blockSignals(False)

        # Get reference points for non-selected variables
        reference_point = self.model_logic.get_reference_points(
            self.ui_components.ref_inputs, self.selected_variables
        )
        generator = cast(BayesianGenerator, self.model_logic.routine.generator)

        logger.debug("Updating plot with selected variables and reference points")

        # Update the plot with the selected variables and reference points
        self.plotting_area.update_plot(
            generator,
            self.selected_variables,
            reference_point,
            self.config_parameters["plot_options"]["show_acq_func"],
            self.config_parameters["plot_options"]["show_samples"],
            self.config_parameters["plot_options"]["show_prior_mean"],
            self.config_parameters["plot_options"]["show_feasibility"],
            self.config_parameters["plot_options"]["n_grid"],
            requires_rebuild,
            interval,
        )

    def update_reference_point_table(self, selected_variables: list[str]):
        """Disable and gray out reference points for selected variables."""
        if not self.model_logic.vocs or not self.ui_components.ref_inputs:
            # vocs or ref_inputs is not yet set; skip processing
            return

        for i, var_name in enumerate(self.model_logic.vocs.variable_names):
            # Get the reference point item from the table
            ref_item = self.ui_components.ref_inputs[i]

            white = Qt.GlobalColor.white
            lightGray = Qt.GlobalColor.lightGray
            black = Qt.GlobalColor.black

            itemIsEditable = Qt.ItemFlag.ItemIsEditable

            if var_name in selected_variables:
                # Disable editing and gray out the background
                ref_item.setFlags(ref_item.flags() & ~Qt.ItemFlags(itemIsEditable))
                ref_item.setBackground(lightGray)
                ref_item.setForeground(white)
            else:
                # Re-enable editing and set background to white
                ref_item.setFlags(ref_item.flags() | Qt.ItemFlags(itemIsEditable))
                ref_item.setBackground(white)
                ref_item.setForeground(black)

        # Force the table to refresh and update its view
        if self.ui_components.reference_table is not None:
            viewport = self.ui_components.reference_table.viewport()
            if viewport is not None:
                viewport.update()



# def test_generator_api():
#     from badger.factory import get_generator

#     _, configs = get_generator('upper_confidence_bound')
#     assert configs['name'] == 'upper_confidence_bound'


def test_env_api():
    from badger.factory import list_env, get_env

    assert len(list_env()) == 2

    _, configs = get_env("test")
    assert configs["name"] == "test"


def test_intf_api():
    from badger.factory import list_intf, get_intf

    assert len(list_intf()) == 1

    _, configs = get_intf("test")
    assert configs["name"] == "test"


# def test_run(mock_config_root):
#     from coolname import generate_slug
#     from badger.log import config_log
#     config_log()
#     from badger.factory import get_generator, get_env
#     from badger.utils import load_config, merge_params, config_list_to_dict
#     from badger.core import normalize_routine, run_routine

#     # Args
#     generator = 'silly'
#     generator_params = '{dimension: 2, max_iter: 10}'
#     env = 'silly'
#     env_params = None
#     config = os.path.join(mock_config_root, 'test.yaml')
#     save = False
#     yes = True
#     verbose = 2

#     # Get env params
#     _, configs_env = get_env(env)

#     # Get generator params
#     _, configs_generator = get_generator(generator)

#     # Normalize the generator and env params
#     params_env = load_config(env_params)
#     params_generator = load_config(generator_params)
#     params_env = merge_params(configs_env['params'], params_env)
#     params_generator = merge_params(configs_generator['params'], params_generator)

#     # Load routine configs
#     configs_routine = load_config(config)

#     # Compose the routine
#     routine = {
#         'name': save or generate_slug(2),
#         'generator': generator,
#         'env': env,
#         'generator_params': params_generator,
#         'env_params': params_env,
#         # env_vranges is an additional info for the normalization
#         # Will be removed after the normalization
#         'env_vranges': config_list_to_dict(configs_env['variables']),
#         'config': configs_routine,
#     }

#     # Sanity check and config normalization
#     routine = normalize_routine(routine)

#     run_routine(routine, yes, save, verbose)



from importlib import resources
import signal
import sys
import time
from PyQt5.QtWidgets import QApplication
from PyQt5.QtGui import QFont, QIcon
from PyQt5 import QtCore
from qdarkstyle import load_stylesheet, LightPalette, DarkPalette

from badger.settings import init_settings
from badger.gui.default.windows.main_window import BadgerMainWindow

import traceback
from badger.errors import BadgerError
from types import TracebackType
from typing import Type, NoReturn

# Fix the scaling issue on multiple monitors w/ different scaling settings
# if sys.platform == 'win32':
#     ctypes.windll.shcore.SetProcessDpiAwareness(1)

if hasattr(QtCore.Qt, "AA_EnableHighDpiScaling"):
    QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)

if hasattr(QtCore.Qt, "AA_UseHighDpiPixmaps"):
    QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

# if hasattr(QtCore.Qt, 'HighDpiScaleFactorRoundingPolicy'):
#     QApplication.setAttribute(
#         QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough)

TIMER = {"time": None}


def on_exit(*args):
    if TIMER["time"] is None:
        print("Press Ctrl/Cmd + C again within 1s to quit Badger")
        TIMER["time"] = time.time()
        return

    TIMER["time"] = None
    print("Goodbye")
    QApplication.quit()


def on_timeout():
    if TIMER["time"] is None:
        return

    now = time.time()
    if (now - TIMER["time"]) > 1:
        TIMER["time"] = None
        print("Timeout, resume the operation...")


def error_handler(
    etype: Type[BaseException], value: BaseException, tb: TracebackType
) -> NoReturn:
    """
    Custom exception handler that formats uncaught exceptions and raises a BadgerError.

    Parameters
    ----------
    etype : Type[BaseException]
        The class of the exception that was raised.
    value : BaseException
        The exception instance.
    tb : TracebackType
        The traceback object associated with the exception.

    Raises
    ------
    BadgerError
        An exception that includes the error title and detailed traceback.
    """
    error_msg = "".join(traceback.format_exception(etype, value, tb))
    error_title = f"{etype.__name__}: {value}"
    raise BadgerError(error_title, error_msg)


def launch_gui(config_path=None):
    sys.excepthook = error_handler

    app = QApplication(sys.argv)

    if config_path is not None:
        config_singleton = init_settings(config_path)
    else:
        config_singleton = init_settings()

    # Set app metainfo
    app.setApplicationName("Badger")
    icon_ref = resources.files(__name__) / "images/icon.png"
    with resources.as_file(icon_ref) as icon_path:
        app.setWindowIcon(QIcon(str(icon_path)))

    font = QFont()
    font.setPixelSize(13)
    # font.setWeight(QFont.DemiBold)
    app.setFont(font)

    # Set up stylesheet
    theme = config_singleton.read_value("BADGER_THEME")
    if theme == "dark":
        app.setStyleSheet(load_stylesheet(palette=DarkPalette))
    elif theme == "light":
        app.setStyleSheet(load_stylesheet(palette=LightPalette))
    else:
        app.setStyleSheet("")

    # Show the main window
    window = BadgerMainWindow()

    # Enable Ctrl + C quit
    signal.signal(signal.SIGINT, on_exit)
    # Let the interpreter run each 0.2 s
    timer = QtCore.QTimer()
    timer.timeout.connect(on_timeout)
    timer.start(200)

    window.show()

    # Show the saving SCORE heads-up
    # QMessageBox.information(
    #        window, 'Heads-up!', 'This might be a good time to save a SCORE.')

    sys.exit(app.exec())



from PyQt5.QtWidgets import QHBoxLayout, QLabel, QLineEdit, QWidget


def labeled_lineedit(name, text, width_name=64, placeholder=None, readonly=True):
    widget = QWidget()
    hbox = QHBoxLayout(widget)
    hbox.setContentsMargins(0, 0, 0, 0)
    label = QLabel(name)
    label.setFixedWidth(width_name)
    widget.edit = edit = QLineEdit()
    if readonly:
        edit.setReadOnly(readonly)
    edit.setText(text)
    if placeholder:
        edit.setPlaceholderText(placeholder)

    hbox.addWidget(label)
    hbox.addWidget(edit, 1)

    return widget



import torch
from badger import environment
from badger.errors import BadgerNoInterfaceError
import time


class Environment(environment.Environment):
    name = "test"
    variables = {f"x{i}": [-1, 1] for i in range(20)}
    observables = ["f", "c"]

    flag: int = 0
    delay: float = 0.0

    def set_variables(self, variable_inputs: dict[str, float]):
        if not self.interface:
            raise BadgerNoInterfaceError

        self.interface.set_values(variable_inputs)
        full_outputs = self.interface.get_values(self.variable_names)

        # Filling up the observations
        x = torch.tensor([full_outputs[f"x{i}"] for i in range(20)])
        self.interface.set_value("f", float((x**2).sum().numpy()))
        self.interface.set_value("c", float((x**2).sum().numpy()))
        time.sleep(self.delay)

    def get_bounds(self, variable_names):
        """
        Returns the bounds of new variables (not already included in env).
        """
        return {name: [-1, 1] for name in variable_names}



import multiprocessing

import pytest
from PyQt5.QtCore import QEventLoop, QTimer


@pytest.fixture(scope="session")
def init_multiprocessing():
    multiprocessing.set_start_method("fork", force=True)


def test_home_page_run_routine(qtbot, init_multiprocessing):
    from badger.archive import save_tmp_run
    from badger.gui.acr.windows.main_window import BadgerMainWindow
    from badger.tests.utils import (
        create_multiobjective_routine,
        create_routine,
        fix_path_issues,
    )

    fix_path_issues()

    main_page = BadgerMainWindow()

    loop = QEventLoop()
    QTimer.singleShot(1000, loop.quit)  # 1000 ms pause
    loop.exec_()

    home_page = main_page.home_page
    # test running routines w high level interface
    routines = [
        create_routine(),
        create_multiobjective_routine(),
        # create_routine_turbo(),
    ]

    for ele in routines:
        tmp_filename = save_tmp_run(ele)
        home_page.current_routine = ele
        home_page.run_monitor.testing = True
        home_page.run_monitor.routine_filename = tmp_filename
        home_page.run_monitor.termination_condition = {
            "tc_idx": 0,
            "max_eval": 3,
        }
        home_page.go_run(-1)
        # start run in a thread and wait some time for it to finish
        home_page.run_monitor.start(True)

        with qtbot.waitSignal(
            home_page.run_monitor.sig_progress, timeout=1000
        ) as blocker:
            pass

        assert blocker.signal_triggered

        # Wait until the run is done
        while home_page.run_monitor.running:
            qtbot.wait(100)

        # assert we get the right result, ie. correct number of samples
        assert len(home_page.run_monitor.routine.data) == 3

    main_page.process_manager.close_proccesses()
